/*! \file
   Provides a powerful generator for pseudo-random numbers. Credits for this 
   solution go to Nathan Reed. See
   http://www.reedbeta.com/blog/2013/01/12/quick-and-easy-gpu-random-numbers-in-d3d11/ */


/*! This function generates a random number using the given integer as seed. The 
   returned random number is distributed uniformly within [0.0,1.0).
  \warning It should only be used on shader model 4.0 and above.*/
float GetRandomNumber(uint Seed){
	// Apply the Wang hash function
	Seed=(Seed^61)^(Seed>>16);
	Seed*=9;
	Seed=Seed^(Seed>>4);
	Seed*=0x27d4eb2d;
	Seed=Seed^(Seed>>15);
	// Go back to a floating point value
	return Seed*pow(0.5f,32);
}


/*! Like GetRandomNumber(uint) but generates its seed from the given two floating 
   point values and implements a fallback for shader model 1.0 to 3.0.*/
float GetRandomNumber(float2 FloatingPointSeed){
#if D3D_VERSION==9
	// Use a cheap linear congruential generator
	return frac(dot(float2(136212.21236f,230875.1235f),FloatingPointSeed));
#elif D3D_VERSION==11
	// Convert the seed to an integer
	uint Seed=asuint(FloatingPointSeed[0])^asuint(FloatingPointSeed[1]);
	return GetRandomNumber(Seed);
#endif
}

/*! Like GetRandomNumber(uint) but generates its seed from the given three floating 
   point values and implements a fallback for shader model 1.0 to 3.0.*/
float GetRandomNumber(float3 FloatingPointSeed){
#if D3D_VERSION==9
	// Use a cheap linear congruential generator
	return frac(dot(float3(136212.21236f,230875.1235f,239108.472f),FloatingPointSeed));
#elif D3D_VERSION==11
	// Convert the seed to an integer
	uint Seed=asuint(FloatingPointSeed[0])^asuint(FloatingPointSeed[1])^asuint(FloatingPointSeed[2]);
	return GetRandomNumber(Seed);
#endif
}

/*! Like GetRandomNumber(uint) but generates its seed from the given four floating 
   point values and implements a fallback for shader model 1.0 to 3.0.*/
float GetRandomNumber(float4 FloatingPointSeed){
#if D3D_VERSION==9
	// Use a cheap linear congruential generator
	return frac(dot(float4(136212.21236f,230875.1235f,239108.472f,489123.434f),FloatingPointSeed));
#elif D3D_VERSION==11
	// Convert the seed to an integer
	uint Seed=asuint(FloatingPointSeed[0])^asuint(FloatingPointSeed[1])^asuint(FloatingPointSeed[2])^asuint(FloatingPointSeed[3]);
	return GetRandomNumber(Seed);
#endif
}
