/*! \file
   This header defines utility functions for Shadow.fx.*/

/*! Given a sampler for the shadow map, the texture coordinate for the center of the 
   kernel and a computed shadow map depth for a point at the same location this 
   function outputs 1.0, if the fragment is in shadow, 0.0f, if the fragment is lit 
   and an intermediate value for partial shadow. It also requires the shadow map 
   size in pixels and its reciproque. It implements PCF with bilinear filtering 
   using a Gaussian kernel with the given standard deviation, which is cut off at 
   the specified radius (both in texels).*/
void ComputePCFShadowIntensity(out float OutShadowIntensity,Texture2D ShadowMap,SamplerState ShadowMapSampler,float2 ShadowMapTexCoord,float FragmentDepth,float4 ShadowMapSize,const float StandardDeviation,const int KernelRadius,float DepthBias){
	float BiasedDepth=mad(FragmentDepth-DepthBias,0.5f,0.5f);
	// Determine the region that needs to be sampled. It is extended by one texel 
	// for bilinear filtering
	const int iFirstSample=-KernelRadius-1;
	const int iLastSample=KernelRadius;
	// Prepare quantities for computation of filter weights
	const float InvVariance=1.0f/(StandardDeviation*StandardDeviation);
	ShadowMapTexCoord+=ShadowMapSize.zw*0.5f;
	float2 TexelCoord=frac(ShadowMapTexCoord*ShadowMapSize.xy);
	OutShadowIntensity=0.0f;
	float TotalWeight=0.0f;
	// Loop over the samples. This code might seem inefficient at first glance but 
	// the HLSL compiler should take good care of it because typically most input 
	// parameters it depends upon are hard-coded literals.
	[unroll] for(int x=iFirstSample;x<=iLastSample;++x){
		// Linearly interpolate between two Gaussian weights to implement bilinear 
		// filtering on top of an image-space Gaussian filtering
		float HorizontalWeight=lerp((x==iLastSample)?0.0f:exp(-0.5f*(x+1)*(x+1)*InvVariance),(x==iFirstSample)?0.0f:exp(-0.5f*x*x*InvVariance),TexelCoord.x);
		[unroll] for(int y=iFirstSample;y<=iLastSample;++y){
			// Obtain the PCF sample
			float2 OffsetTexCoord=ShadowMapTexCoord+ShadowMapSize.zw*float2(x,y);
			float SampledShadowMapDepth=ShadowMap.Sample(ShadowMapSampler,OffsetTexCoord);
			float Summand=(SampledShadowMapDepth<BiasedDepth)?1.0f:0.0f;
			// Compute the vertical weight in analogy to the horizontal weight
			float VerticalWeight=lerp((y==iLastSample)?0.0f:exp(-0.5f*(y+1)*(y+1)*InvVariance),(y==iFirstSample)?0.0f:exp(-0.5f*y*y*InvVariance),TexelCoord.y);
			float Weight=HorizontalWeight*VerticalWeight;
			OutShadowIntensity+=Weight*Summand;
			TotalWeight+=Weight;
		}
	}
	OutShadowIntensity/=TotalWeight;
}


/*! Variant of ComputePCFShadowIntensity() which makes use of SampleCmp() to improve 
   the performance.*/
void ComputePCFShadowIntensityBilinear(out float OutShadowIntensity,Texture2D<float> ShadowMap,SamplerComparisonState ShadowMapSampler,float2 ShadowMapTexCoord,float FragmentDepth,float4 ShadowMapSize,const float StandardDeviation,const int KernelRadius,float DepthBias){
	float BiasedDepth=mad(FragmentDepth-DepthBias,0.5f,0.5f);
	// Determine the region that needs to be sampled. Conveniently, this is always a 
	// region with an even edge-length because the kernel has an odd edge-length and 
	// one additional sample is needed for bilinear interpolation.
	const int iFirstSample=-KernelRadius-1;
	const int iLastSample=KernelRadius;
	// Prepare quantities for computation of filter weights. In particular we need a 
	// shadow map texture coordinate located exactly in the center of a texel i.e. 
	// sampling only one value.
	const float InvVariance=1.0f/(StandardDeviation*StandardDeviation);
	ShadowMapTexCoord+=ShadowMapSize.zw*0.5f;
	float2 TexelCoord=frac(ShadowMapTexCoord*ShadowMapSize.xy);
	ShadowMapTexCoord-=ShadowMapSize.zw*TexelCoord-ShadowMapSize.zw*0.5f;
	OutShadowIntensity=0.0f;
	float TotalWeight=0.0f;
	// Loop over the samples always taking two steps at a time. We use SampleCmp() 
	// to get four samples at once without changing the weighting.
	[unroll] for(int x=iFirstSample;x<=iLastSample;x+=2){
		// Compute weights for the left and right boundary of the texel
		int LeftX=x;
		float LeftWeight=lerp(exp(-0.5f*(LeftX+1)*(LeftX+1)*InvVariance),(LeftX==iFirstSample)?0.0f:exp(-0.5f*LeftX*LeftX*InvVariance),TexelCoord.x);
		int RightX=x+1;
		float RightWeight=lerp((RightX==iLastSample)?0.0f:exp(-0.5f*(RightX+1)*(RightX+1)*InvVariance),exp(-0.5f*RightX*RightX*InvVariance),TexelCoord.x);
		// Compute the relation between these weights to pipe it into SampleCmp() as 
		// texture coordinate
		float HorizontalWeight=LeftWeight+RightWeight;
		float RelativeHorizontalWeight=RightWeight/HorizontalWeight;
		[unroll] for(int y=iFirstSample;y<=iLastSample;y+=2){
			// Compute vertical weights in analogy to the horizontal weights
			int TopY=y;
			float TopWeight=lerp(exp(-0.5f*(TopY+1)*(TopY+1)*InvVariance),(TopY==iFirstSample)?0.0f:exp(-0.5f*TopY*TopY*InvVariance),TexelCoord.y);
			int BottomY=y+1;
			float BottomWeight=lerp((BottomY==iLastSample)?0.0f:exp(-0.5f*(BottomY+1)*(BottomY+1)*InvVariance),exp(-0.5f*BottomY*BottomY*InvVariance),TexelCoord.y);
			float VerticalWeight=TopWeight+BottomWeight;
			float RelativeVerticalWeight=BottomWeight/VerticalWeight;
			// Take a comparison sample using the relative weights as intra-texel 
			// coordinate
			float2 OffsetTexCoord=ShadowMapTexCoord+ShadowMapSize.zw*float2(x+RelativeHorizontalWeight,y+RelativeVerticalWeight);
			float Summand=ShadowMap.SampleCmpLevelZero(ShadowMapSampler,OffsetTexCoord,BiasedDepth);
			float Weight=HorizontalWeight*VerticalWeight;
			OutShadowIntensity+=Weight*Summand;
			TotalWeight+=Weight;
		}
	}
	OutShadowIntensity/=TotalWeight;
}


/*! Solves the polynomial Coefficients[0]+Coefficients[1]*x+Coefficients[2]*x^2 for 
   x and returns a vector consisting of the two solutions. The returned solutions 
   are sorted. Behavior is undefined if the polynomial has no roots. A root with 
   multiplicity is returned twice.*/
float2 GetRoots(float3 Coefficients){
	float Scaling=1.0f/Coefficients[2];
	float p=Coefficients[1]*Scaling;
	float q=Coefficients[0]*Scaling;
	float D=p*p*0.25f-q;
	float r=sqrt(D);
	return float2(-0.5f*p-r,-0.5f*p+r);
}


/*! Computes the determinant of the Hankel matrix induced by the given four moments 
   robustly by using a Cholesky decomposition.*/
float GetHankelDeterminant(float4 Moments){
	float L21D11=mad(-Moments[0],Moments[1],Moments[2]);
	float D11=mad(-Moments[0],Moments[0], Moments[1]);
	float SquaredDepthVariance=mad(-Moments[1],Moments[1], Moments[3]);
	return dot(float2(SquaredDepthVariance,-L21D11),float2(D11,L21D11));
}


/*! Computes the determinant of the Hankel matrix induced by the given four moments 
   divided by the variance. This coincides with the amount by which the fourth 
   moment is greater than the minimal possible value.*/
float GetFourthMomentOffset(float4 Moments){
	float D11=mad(-Moments[0],Moments[0],Moments[1]);
	float L21D11=mad(-Moments[0],Moments[1],Moments[2]);
	float L21=L21D11/D11;
	float SquaredDepthVariance=mad(-Moments[1],Moments[1],Moments[3]);
	return mad(-L21D11,L21,SquaredDepthVariance);
}
