/*! \file
   This header defines advanced utility functions to deal with complex matrices and 
   complex polynomials. It is specifically used for computing lower bounds to CDFs 
   in trigonometric moment problems.*/

#include "ComplexAlgebra.fx"
   
/*! @{
   This function evaluates the standard Hermitian form of two complex vectors.*/
float2 Dot(float3x2 LHS,float3x2 RHS){
	return float2(
		dot(transpose(LHS)[0],transpose(RHS)[0])+dot(transpose(LHS)[1],transpose(RHS)[1]),
		dot(transpose(LHS)[0],transpose(RHS)[1])-dot(transpose(LHS)[1],transpose(RHS)[0]));
}
float2 Dot(float4x2 LHS,float4x2 RHS){
	return float2(
		dot(transpose(LHS)[0],transpose(RHS)[0])+dot(transpose(LHS)[1],transpose(RHS)[1]),
		dot(transpose(LHS)[0],transpose(RHS)[1])-dot(transpose(LHS)[1],transpose(RHS)[0]));
}
float2 Dot(float2 LHS[5],float2 RHS[5]){
	float2 Result=float2(0.0f,0.0f);
	[unroll] for(int i=0;i!=5;++i){
		Result+=Multiply(LHS[i],RHS[i]);
	}
	return Result;
}
//!@}


/*! @{
   Multiplies a complex valued vector by a complex valued matrix. The vectors are 
   given as matrices with two columns where the first column holds the real part and 
   the second column holds the imaginary parth. The matrix is given as two real 
   matrices, one for the real part, the other for the imaginary part. Matrix 
   multiplication is done from the right.*/
float3x2 Multiply(float3x3 MatrixRealPart,float3x3 MatrixImaginaryPart,float3x2 RHSVector){
	float3x2 Result;
	Result._m00_m10_m20=mul(MatrixRealPart,RHSVector._m00_m10_m20)-mul(MatrixImaginaryPart,RHSVector._m01_m11_m21);
	Result._m01_m11_m21=mul(MatrixRealPart,RHSVector._m01_m11_m21)+mul(MatrixImaginaryPart,RHSVector._m00_m10_m20);
	return Result;
}
float4x2 Multiply(float4x4 MatrixRealPart,float4x4 MatrixImaginaryPart,float4x2 RHSVector){
	float4x2 Result;
	Result._m00_m10_m20_m30=mul(MatrixRealPart,RHSVector._m00_m10_m20_m30)-mul(MatrixImaginaryPart,RHSVector._m01_m11_m21_m31);
	Result._m01_m11_m21_m31=mul(MatrixRealPart,RHSVector._m01_m11_m21_m31)+mul(MatrixImaginaryPart,RHSVector._m00_m10_m20_m30);
	return Result;
}
void Multiply(out float2 OutProduct[5],float MatrixRealPart[5][5],float MatrixImaginaryPart[5][5],float2 RHSVector[5]){
	[unroll] for(int i=0;i!=5;++i){
		OutProduct[i]=float2(0.0f,0.0f);
		[unroll] for(int j=0;j!=5;++j){
			OutProduct[i]+=Multiply(float2(MatrixRealPart[i][j],MatrixImaginaryPart[i][j]),RHSVector[j]);
		}
	}
}
//!@}



/*! This function outputs the two roots of the specified quadratic polynomial. The 
   complex number given by pCoefficient[j] is used as coefficient for the j-th 
   power. No guarantees are made concerning the order of the output roots. Roots 
   with multiplicity are output repeatedly.*/
float2x2 GetRoots(float3x2 pCoefficient){
	float2 aInv=Reciprocal(pCoefficient[2]);
	float2 p=Multiply(pCoefficient[1],aInv);
	float2 q=Multiply(pCoefficient[0],aInv);
	float2 Discriminant=0.25f*Square(p)-q;
	float2 Root=SquareRoot(Discriminant);
	float2 pHalf=-0.5f*p;
	float2x2 pRoot;
	pRoot[0]=pHalf-Root;
	pRoot[1]=pHalf+Root;
	return pRoot;
}

/*! Like float2x2 GetRoots(float3x2) but for cubic polynomials.*/
float3x2 GetRoots(float4x2 pCoefficient){
	// This implementation is based upon the algorithm and notions on Wikipedia:
	// http://en.wikipedia.org/wiki/Cubic_polynomial#General_formula_for_roots
	float2 bb=Square(pCoefficient[2]);
	float2 ac=Multiply(pCoefficient[1],pCoefficient[3]);
	float2 Delta0=bb-3.0f*ac;
	float2 bbb=Multiply(bb,pCoefficient[2]);
	float2 abc=Multiply(ac,pCoefficient[2]);
	float2 aad=Multiply(Multiply(pCoefficient[3],pCoefficient[3]),pCoefficient[0]);
	float2 Delta1=2.0f*bbb-9.0f*abc+27.0f*aad;
	float2 C=CubicRoot((Delta1+SquareRoot(Multiply(Delta1,Delta1)-4.0f*Cube(Delta0)))*0.5f);
	float2 aInv=Reciprocal(pCoefficient[3]);
	float2 Delta0OverC=Divide(Delta0,C);
	float3x2 pRoot;
	pRoot[0]=-(1.0f/3.0f)*Multiply(aInv,pCoefficient[2]+C+Delta0OverC);
	pRoot[1]=-(1.0f/3.0f)*Multiply(aInv,pCoefficient[2]+Multiply(float2(-1.0f, sqrt(3.0f))/2.0f,C)+Multiply(float2(-1.0f,-sqrt(3.0f))/2.0f,Delta0OverC));
	pRoot[2]=-(1.0f/3.0f)*Multiply(aInv,pCoefficient[2]+Multiply(float2(-1.0f,-sqrt(3.0f))/2.0f,C)+Multiply(float2(-1.0f, sqrt(3.0f))/2.0f,Delta0OverC));
	return pRoot;
}

/*! Like float2x2 GetRoots(float3x2) but for quartic polynomials.*/
float4x2 GetRoots(float2 pCoefficient[5]){
	// This implementation is based upon the algorithm and notions on Wikipedia:
	// http://en.wikipedia.org/wiki/Quartic_equation#General_formula_for_roots
	float2 cc=Square(pCoefficient[2]);
	float2 bd=Multiply(pCoefficient[1],pCoefficient[3]);
	float2 ae=Multiply(pCoefficient[0],pCoefficient[4]);
	float2 Delta0=cc-3.0f*bd+12.0f*ae;
	
	float2 ccc=Multiply(cc,pCoefficient[2]);
	float2 bcd=Multiply(bd,pCoefficient[2]);
	float2 bb=Square(pCoefficient[3]);
	float2 bbe=Multiply(bb,pCoefficient[0]);
	float2 add=Multiply(Square(pCoefficient[1]),pCoefficient[4]);
	float2 ace=Multiply(ae,pCoefficient[2]);
	float2 Delta1=2.0f*ccc-9.0f*bcd+27.0f*bbe+27.0f*add-72.0f*ace;

	float2 Q=CubicRoot((Delta1+SquareRoot(Square(Delta1)-4.0f*Cube(Delta0)))*0.5f);
	
	float2 ac=Multiply(pCoefficient[2],pCoefficient[4]);
	float2 aa=Square(pCoefficient[4]);
	float2 p=Divide(8.0f*ac-3.0f*bb,8.0f*aa);
	float2 aaa=Multiply(aa,pCoefficient[4]);
	float2 aad=Multiply(aa,pCoefficient[1]);
	float2 abc=Multiply(pCoefficient[2],Multiply(pCoefficient[3],pCoefficient[4]));
	float2 bbb=Multiply(bb,pCoefficient[3]);
	float2 q=Divide(bbb-4.0f*abc+8.0f*aad,8.0f*aaa);

	float2 aInv=Reciprocal(pCoefficient[4]);
	float2 S=0.5f*SquareRoot((-2.0f/3.0f)*p+(1.0f/3.0f)*Multiply(aInv,Q+Divide(Delta0,Q)));

	float2 bOver4a=0.25f*Multiply(pCoefficient[3],aInv);
	float2 SS4Plus2p=4.0f*Square(S)+2.0f*p;
	float2 qOverS=Divide(q,S);
	float2 FirstSquareRoot=SquareRoot(-SS4Plus2p+qOverS);
	float2 SecondSquareRoot=SquareRoot(-SS4Plus2p-qOverS);

	float4x2 pRoot;
	pRoot[0]=-bOver4a-S+0.5f*FirstSquareRoot;
	pRoot[1]=-bOver4a-S-0.5f*FirstSquareRoot;
	pRoot[2]=-bOver4a+S+0.5f*SecondSquareRoot;
	pRoot[3]=-bOver4a+S-0.5f*SecondSquareRoot;
	return pRoot;
}


/*! This function computes all roots of the depressed cubic equation x^3+p*x+q=0 
   under the assumption, that it has three real roots (counting multiplicity).*/
float3 GetRoots(float p,float q){
	// Use Vieta's substitution
	float2 w=float2(-0.5f*q,sqrt(-(0.25f*q*q+p*p*p/27.0f)));
	// Compute the third roots
	float2 pThirdRoot[3];
	pThirdRoot[0]=CubicRoot(w);
	pThirdRoot[1]=Multiply(pThirdRoot[0],float2(-0.5f,0.5f*sqrt(3.0f)));
	pThirdRoot[2]=Multiply(pThirdRoot[0],float2(-0.5f,-0.5f*sqrt(3.0f)));
	// Revert the substitution and extract the real part
	float3 pRoot;
	[unroll] for(int i=0;i!=3;++i){
		pRoot[i]=pThirdRoot[i].x-p*(1.0f/3.0f)*Reciprocal(pThirdRoot[i]).x;
	}
	return pRoot;
}


/*! @{
   Given an array consisting of normalized trigonometric moments (starting with the 
   first moment) this function outputs the real part and imaginary part of the 
   inverse of the corresponding Toeplitz matrix.
  \warning Whenever possible SolveToeplitzSystem() should be preferred due to its 
		   better numerical stability.*/
void GetToeplitzInverse(out float3x3 OutToeplitzInverseReal,out float3x3 OutToeplitzInverseImaginary,float2 pTrigonometricMoment[2]){
	float2 b1=pTrigonometricMoment[0];
	float2 b2=pTrigonometricMoment[1];
	float m00=1.0f-dot(b1,b1);
	float2 m10=-b1+Multiply(Conjugate(b1),b2);
	float m11=1.0f-dot(b2,b2);
	float2 m20=Multiply(b1,b1)-b2;
	OutToeplitzInverseReal=float3x3(
		m00.x,m10.x,m20.x,
		m10.x,m11.x,m10.x,
		m20.x,m10.x,m00.x
	);
	OutToeplitzInverseImaginary=float3x3(
		  0.0f,-m10.y,-m20.y,
		 m10.y,  0.0f,-m10.y,
		 m20.y, m10.y,  0.0f
	);
	// Compute the real part of the (non-Hermitian) dot product of the first row of 
	// the scaled inverse with the first column of the Toeplitz matrix
	float Scaling=dot(OutToeplitzInverseReal[0],float3(1.0f,pTrigonometricMoment[0].x,pTrigonometricMoment[1].x))-dot(OutToeplitzInverseImaginary[0],float3(0.0f,pTrigonometricMoment[0].y,pTrigonometricMoment[1].y));
	// It should be one, so we need to divide by that
	OutToeplitzInverseReal/=Scaling;
	OutToeplitzInverseImaginary/=Scaling;
}


/*! This function solves a linear system of the form Toeplitz*Result=RHS where 
   Toeplitz is a Hermitian Toeplitz matrix with one on the diagonal and the given 
   trigonometric moments below that in the first column. RHS is a complex row-vector 
   where the first entry is one and the other two are given. Note that the Toeplitz 
   matrix has to be positive definite. It returns Result.*/
float3x2 SolveToeplitzSystem(float2 TrigonometricMoment1,float2 TrigonometricMoment2,float2 RHS1,float2 RHS2){
	float2 b1=TrigonometricMoment1;
	float2 b2=TrigonometricMoment2;
	// Construct a Cholesky decomposition of the Toeplitz matrix
	float D11=1.0f-dot(b1,b1);
	float InvD11=1.0f/D11;
	float2 L21=(b1-Multiply(Conjugate(b1),b2))*InvD11;
	float D22=1.0f-dot(b2,b2)-dot(L21,L21)*D11;
	// Perform forward and backward substitution to solve the linear system
	float3x2 c;
	c[0]=float2(1.0f,0.0f);
	c[1]=RHS1;
	c[2]=RHS2;
	// Forward substitution to solve L*c1=bz
	c[1]-=b1;
	c[2]-=b2+Multiply(L21,c[1]);
	// Scaling to solve D*c2=c1
	c[1]*=InvD11;
	c[2]*=1.0f/D22;
	// Backward substitution to solve L^H*c3=c2
	c[1]-=Multiply(Conjugate(L21),c[2]);
	c[0]-=Multiply(Conjugate(b1),c[1])+Multiply(Conjugate(b2),c[2]);
	return c;
}


/*! This function returns the determinant of the Toeplitz matrix arising from the 
   given normalized, trigonometric moments.*/
float GetDeterminant(float2 pTrigonometricMoment[2]){
	return 1.0f-(2.0f*dot(pTrigonometricMoment[0],pTrigonometricMoment[0])+dot(pTrigonometricMoment[1],pTrigonometricMoment[1]))+2.0f*Multiply(Square(pTrigonometricMoment[0]),Conjugate(pTrigonometricMoment[1])).x;
}
