/*! This function sets AdditiveColor either to opaque black, if AdditiveColorUsed is 
   false or to SourceColor, if AdditiveColorUsed is true.*/
void ChooseAdditiveColor(out float4 OutAdditiveColor,bool AdditiveLightingEnabled,float4 SourceColor){
	OutAdditiveColor=AdditiveLightingEnabled?SourceColor:float4(0.0f,0.0f,0.0f,1.0f);
}

/*! This function performs a multiplication of two colors.*/
void MultiplyColors(out float4 OutColor,float4 Argument1,float4 Argument2){
	OutColor=Argument1*Argument2;
}

/*! This function performs a multiplication of two colors, followed by the 
   multiplication of the factor 2.0.*/
void MultiplyColorsX2(out float4 OutColor,float4 Argument1,float4 Argument2){
	OutColor.rgb=Argument1.rgb*Argument2.rgb*2.0f;
	OutColor.a=Argument1.a*Argument2.a;
}

/*! This function performs adds together two colors and a constant value and 
   multiplies the result by a given value.*/
void AddColors(out float4 OutColor,float4 Argument1,float4 Argument2){
	OutColor.rgb=Argument1.rgb+Argument2.rgb;
	OutColor.a=Argument1.a*Argument2.a;
}

/*! This function subtracts two colors.*/
void SubtractColors(out float4 OutColor,float4 Argument1,float4 Argument2){
	OutColor.rgb=Argument1.rgb-Argument2.rgb;
	OutColor.a=Argument1.a*Argument2.a;
}

/*! This function adds together argument one and the product of argument 2 and 
   the inverse of argument 1.*/
void AddColorsSmooth(out float4 OutColor,float4 Argument1,float4 Argument2){
	OutColor.rgb=Argument1.rgb+Argument2.rgb*(float3(1.0f,1.0f,1.0f)-Argument1.rgb);
	OutColor.a=Argument1.a*Argument2.a;
}

/*! This function lerps between both arguments using the alpha of the second argument.*/
void AlphaBlendColors(out float4 OutColor,float4 Argument1,float4 Argument2){
	OutColor.rgb=Argument1.rgb+(Argument2.rgb-Argument1.rgb)*Argument2.a;
	OutColor.a=Argument1.a*Argument2.a;
}


/*! This function computes the 3D dot product of Argument1 and Argument2 and 
   replicates it to all color channels of the output.*/
void DotProduct3(out float4 OutColor,float4 Argument1,float4 Argument2){
	OutColor.rgb=dot(Argument1.rgb*2.0f-float3(1.0f,1.0f,1.0f),Argument2.rgb*2.0f-float3(1.0f,1.0f,1.0f)).rrr;
	OutColor.a=Argument1.a*Argument2.a;
}


/*! This function outputs Argument1 ignoring Argument2.*/
void UseFirstColor(out float4 OutColor,float4 Argument1,float4 Argument2){
	OutColor=Argument1;
}


/*! This function outputs Argument2 ignoring Argument1.*/
void UseSecondColor(out float4 OutColor,float4 Argument1,float4 Argument2){
	OutColor=Argument2;
}


/*! This function applies a given alpha value to a color (overwriting the old alpha) 
   and outputs the result.*/
void ApplyAlpha(out float4 OutColorAlpha,float4 Color,float Alpha){
	OutColorAlpha.rgb=Color.rgb;
	OutColorAlpha.a=Alpha;
}


/*! This function appends the alpha of a given RGBA-color to a given RGB-color and 
   outputs the result.*/
void AppendAlpha(out float4 OutColorAlpha,float3 RGBColor,float4 RGBAColor){
	OutColorAlpha.rgb=RGBColor;
	OutColorAlpha.a=RGBAColor.a;
}

/*! This function extracts the alpha value from a given color and outputs it as result.*/
void GetAlpha(out float OutAlpha,float4 Color){
	OutAlpha=Color.a;
}

/*! This function outputs a negative value (indicating that a pixel should be 
   discarded if it is forwarded to the right output) if and only if the given alpha 
   value is less than the threshold.*/
void DiscardTransparentPixel(out bool OutDiscardPixel,float Alpha,float AlphaThreshold=0.03f){
	OutDiscardPixel=(Alpha<AlphaThreshold);
}

/*! This function brings the components of the given vector from the range from 0 to 
   1 into the range from -1 to 1 using a linear transform. It can be used e.g. to 
   decode normal maps.*/
void DecodeBiasedVector(out float3 OutDirectionVector,float3 BiasedVector){
	OutDirectionVector=BiasedVector*2.0f-float3(1.0f,1.0f,1.0f);
}


/*! This function lerps two colors component wise using a given scalar as lerp 
   factor. For LerpScalar==0.0 Color1 is returned, for LerpScalar==1.0 Color2 is 
   returned.*/
void LerpColors(out float4 OutLerpColor,float4 Color1,float4 Color2,float LerpScalar){
	OutLerpColor=lerp(Color1,Color2,LerpScalar);
}


/*! Given an RGBA vector providing a color and an alpha value, this function outputs 
   its RGB color part separated from the alpha part.*/
void SplitColorAndAlpha(out float3 OutColor,out float OutAlpha,float4 RGBAColor){
	OutColor.rgb=RGBAColor.rgb;
	OutAlpha=RGBAColor.a;
}
