/*! This function computes the normalized world space light direction, which results 
   from a point light lighting a specified surface element. In addition it outputs 
   some intermediate data, which can be useful for further computations.
  \param IdealIncomingWorldSpaceLightDirection The normalized world space direction 
         vector defining the direction of the light meeting the specified surface 
		 element.
  \param InvLightDistance The reciproque of the distance from the light source to 
         the surface element.
  \param WorldSpaceLightPosition The position of the point light in world space.
  \param WorldSpacePosition The position of the surface element in world space.
  \note This function is useful for any point light independent of its angular light 
        distribution.*/
void ComputePointLightDirection(out float3 IdealIncomingWorldSpaceLightDirection,out float InvLightDistance,float3 WorldSpaceLightPosition,float3 WorldSpacePosition){
	float3 UnnormalizedLightDirection=WorldSpacePosition-WorldSpaceLightPosition;
	float LightDistanceSq=dot(UnnormalizedLightDirection,UnnormalizedLightDirection);
	InvLightDistance=rsqrt(LightDistanceSq);
	IdealIncomingWorldSpaceLightDirection=UnnormalizedLightDirection*InvLightDistance;
}

/*! This function computes the irradiance received from a point light source.
   \param IdealIncomingIrradiance The irradiance received by the surface element of 
          interest.
   \param UnitSphereIrradiance The irradiance, which would be received by a surface 
          element on the unit sphere (radius one meter), if this surface element 
		  happens to be on the straight from the light source to the surface element 
		  of interest.
  \param InvLightDistance The reciproque of the distance from the light source to 
         the surface element.
  \note This function is useful for any point light independent of its angular light 
        distribution. You only need to define UnitSphereIrradiance correctly.*/
void ComputePointLightIrradiance(out float3 IdealIncomingIrradiance,float3 UnitSphereIrradiance,float InvLightDistance){
	IdealIncomingIrradiance=UnitSphereIrradiance*(InvLightDistance*InvLightDistance);
}


/*! This function computes the irradiance received by the unit sphere from an 
   isotropic point light source with the given power in the center of the sphere.
   Note that this irradiance is the same at every point of the unit sphere.*/
void ComputePointLightUnitSphereIrradiance(out float3 UnitSphereIrradiance,float3 Power){
	// Irradiance is power per area, so dividing by the area of the unit sphere 
	// (4*pi) does the trick
	UnitSphereIrradiance=0.07957747154594766788444188168626f*Power;
}


/*! This function computes coefficients for a linear function, which maps a cosine 
   value to an attenuation factor. The gradient is written to 
   ConeAngleCoefficients.x, the y-intercept is written to y. The linear function 
   maps cos(InnerConeAngle) to 1.0 and cos(OuterConeAngle) to 0.0. Beyond that this 
   function computes the maximal irradiance received from the spot light at unit 
   distance using the power and the angles.*/
void ComputeSpotConeCoefficients(out float2 ConeAngleCoefficients,out float3 MaxUnitSphereIrradiance,float InnerConeAngle,float OuterConeAngle,float3 Power){
	// Compute the two cosine values
	float CosInnerCone=cos(InnerConeAngle);
	float CosOuterCone=cos(OuterConeAngle);
	// Solve m*CosInnerCone+b=1 and m*CosOuterCone+b=0 for m and b
	// m*(CosInnerCone-CosOuterCone)=1 => m=1/(CosInnerCone-CosOuterCone)
	// b=-m*CosOuterCone
	ConeAngleCoefficients.x=1.0f/(CosInnerCone-CosOuterCone);
	ConeAngleCoefficients.y=-ConeAngleCoefficients.x*CosOuterCone;
	// Compute the maximal unit sphere irradiance (see 
	// Documents/PhysicsBasedSpotLight.lyx for the integrals leading to this 
	// formula)
	const float PI=3.1415926535897932384626433832795f;
	MaxUnitSphereIrradiance=Power/(PI*(2.0f-CosInnerCone-CosOuterCone));
}

/*! This function computes the attenuation of a spot light, which is caused by the 
   angle between the given incoming light direction and the orientation of the 
   light. The output value is in the range from 0.0 to 1.0.*/
void ComputeSpotConeAttenuation(out float SpotConeAttenuation,float3 IdealIncomingWorldSpaceLightDirection,float3 WorldSpaceConeAxisDirection,float2 ConeAngleCoefficients){
	float CosSpotAngle=dot(IdealIncomingWorldSpaceLightDirection,WorldSpaceConeAxisDirection);
	SpotConeAttenuation=saturate(ConeAngleCoefficients.x*CosSpotAngle+ConeAngleCoefficients.y);
}

/*! This function computes the irradiance received by the surface element on the 
   unit sphere surrounding the spot light, which lies on the straight from the spot 
   light to the surface element of interest.*/
void ComputeSpotLightUnitSphereIrradiance(out float3 UnitSphereIrradiance,float3 MaxUnitSphereIrradiance,float SpotConeAttenuation){
	UnitSphereIrradiance=MaxUnitSphereIrradiance*SpotConeAttenuation;
}
