/*! This function multiplies a radiance value point-wise by a specified factor and 
   outputs the result.*/
void ScaleRadiance(out float3 OutRadiance,float3 Factor,float3 Radiance){
	OutRadiance=Factor*Radiance;
}

/*! This function multiplies a given first radiance point-wise by a given factor and 
   adds a given second radiance to the result, which is then output.*/
void ScaleAndAddRadiance(out float3 OutRadiance,float Factor,float3 Radiance1,float3 Radiance2){
	OutRadiance=Factor*Radiance1+Radiance2;
}

/*! This function adds up the given two radiance values and outputs the result.*/
void AddRadiance(out float3 OutRadiance,float3 Radiance1,float3 Radiance2){
	OutRadiance=Radiance1+Radiance2;
}

/*! This function computes the specular lighting color for a colored light source.*/
void ComputeSpecularLightingColor(out float3 SpecularLightingColor,float3 UsedSpecularColor,float UnraisedSpecularIntensity,float SpecularPower){
	SpecularLightingColor=UsedSpecularColor*pow(UnraisedSpecularIntensity,SpecularPower);
}

/*! Given an irradiance this function scales it according to the given shadow 
   intensity, i.e. it multiplies it by 1.0 minus the shadow intensity.*/
void ApplyShadowIntensity(out float3 OutShadowedIrradiance,float3 Irradiance,float ShadowIntensity){
	OutShadowedIrradiance=Irradiance*(1.0f-ShadowIntensity);
}

/*! This function simply outputs the given input vector unmodified. It is used to 
   enforce computations relying on the world space position to be performed in the 
   fragment shader.*/
void CopyWorldSpacePosition(out float3 OutWorldSpacePosition,float3 WorldSpacePosition){
	OutWorldSpacePosition=WorldSpacePosition;
}

/*! This function normalizes a normal vector (i.e. it ensures that it has unit 
   length).*/
void NormalizeNormal(out float3 NormalizedNormal,float3 Normal){
	NormalizedNormal=normalize(Normal);
}

/*! This function computes the normalized vector pointing from the surface element 
   to the camera position in world space.*/
void ComputeWorldSpaceViewDirection(out float3 WorldSpaceViewDirection,float3 WorldSpacePosition,float3 WorldSpaceCameraPosition){
	WorldSpaceViewDirection=normalize(WorldSpaceCameraPosition-WorldSpacePosition);
}

/*! This function computes the normalized vector, which results from reflecting the 
   world space camera to fragment vector on the plane defined by the world space 
   normal. The given world space normal has to be normalized.*/
void ComputeWorldSpaceViewReflectionDirection(out float3 WorldSpaceViewReflectionDirection,float3 WorldSpaceViewDirection,float3 WorldSpaceNormal){
	WorldSpaceViewReflectionDirection=-WorldSpaceViewDirection-2.0f*dot(-WorldSpaceViewDirection,WorldSpaceNormal)*WorldSpaceNormal;
}

/*! This function evaluates the proportionality factor of the cosine lobe BRDF, 
   which is responsible for energy conservation.*/
void ComputeCosineLobeFactor(out float CosineLobeFactor,float CosineLobePower){
	const float INV_2PI=0.15915494309189533576888376337251f;
	const float INV_PI=0.31830988618379067153776752674503;
	CosineLobeFactor=CosineLobePower*INV_2PI+INV_PI;
}

/*! This function evaluates the cosine lobe BRDF for the specified directions and 
   returns the resulting BRDF coefficient. CosineLobeFactor is the output of 
   ComputeCosineLobeFactor().*/
void ComputeCosineLobeBRDF(out float BRDFCoefficient,float CosineLobeFactor,float CosineLobePower,float3 WorldSpaceViewReflectionDirection,float3 IdealIncomingWorldSpaceLightDirection){
	BRDFCoefficient=pow(max(0.0f,-dot(WorldSpaceViewReflectionDirection,IdealIncomingWorldSpaceLightDirection)),CosineLobePower)*CosineLobeFactor;
}


/*! Computes the various constant quantities needed for evaluation of the Torrance 
   Sparrow BRDF from quantities which would typically function as input to cosine 
   lobe or Blinn-Phong.
  \sa ComputeTorranceSparrowBRDF */
void ComputeTorranceSparrowConstants(out float TorranceSparrowFactor,out float MinFresnelTerm,out float NormalDistributionPower,out float SchlickSmithLerpFactor,float CosineLobePower){
	const float PI=3.1415926535897932384626433832795f;
	MinFresnelTerm=0.5f;
	NormalDistributionPower=CosineLobePower;
	TorranceSparrowFactor=(NormalDistributionPower+2.0f)/(8.0f*PI);
	SchlickSmithLerpFactor=rsqrt(0.25f*PI*NormalDistributionPower+0.5f*PI);
}

/*! This function evaluates an optimized variant of the Torrance Sparrow BRDF for 
   the specified directions and returns the resulting BRDF coefficient. The 
   implementation goes back to the talk SIGGRAPH 2012 presentation "Calibrating 
   Lighting and Materials in Far Cry 3" by Steve McAuley.
  \sa ComputeTorranceSparrowConstants() */
void ComputeTorranceSparrowBRDF(out float BRDFCoefficient,float TorranceSparrowFactor,float MinFresnelTerm,float NormalDistributionPower,float SchlickSmithLerpFactor,float3 WorldSpaceViewDirection,float3 IdealIncomingWorldSpaceLightDirection,float3 WorldSpaceNormal){
	// Construct the half vector
	float3 WorldSpaceHalfVector=normalize(WorldSpaceViewDirection-IdealIncomingWorldSpaceLightDirection);
	// Compute all relevant dot products and do not let them get negative
	float NormalDotLight=saturate(dot(WorldSpaceNormal,-IdealIncomingWorldSpaceLightDirection));
	float NormalDotView=saturate(dot(WorldSpaceNormal,WorldSpaceViewDirection));
	float NormalDotHalf=saturate(dot(WorldSpaceNormal,WorldSpaceHalfVector));
	float HalfDotView=saturate(dot(WorldSpaceHalfVector,WorldSpaceViewDirection));
	// Compute the normal distribution term
	float NormalDistributionTerm=pow(NormalDotHalf,NormalDistributionPower);
	// Approximate the Fresnel term using a spherical Gaussian approximation
	float FresnelTerm=lerp(MinFresnelTerm,1.0f,exp(-6.0f*HalfDotView));
	// Compute the visibility term which includes the geometry term
	float VisibilityTerm=1.0f/(lerp(NormalDotLight,1.0f,SchlickSmithLerpFactor)*lerp(NormalDotView,1.0f,SchlickSmithLerpFactor));
	// Put it all together
	BRDFCoefficient=max(0.0f,TorranceSparrowFactor*FresnelTerm*NormalDistributionTerm*VisibilityTerm);
}

/*! This function computes the negated dot product between the given two vectors and 
   clamps it to the non-negative range. It is meant to be used to compute the cosine 
   factor, which is needed to regard the orientation of a surface element when 
   computing incoming irradiance.
  \param IrradianceCosineFactor The cosine factor described above.
  \param IdealIncomingWorldSpaceLightDirection The normalized direction vector 
         defining the direction from which the light is coming in from the light 
		 source. It is given in world space and points in movement direction of the 
		 light.
  \param WorldSpaceNormal The normalized normal vector of the surface element of 
         interest in world space.*/
void ComputeIrradianceCosineFactor(out float IrradianceCosineFactor,float3 IdealIncomingWorldSpaceLightDirection,float3 WorldSpaceNormal){
	IrradianceCosineFactor=max(0.0f,-dot(IdealIncomingWorldSpaceLightDirection,WorldSpaceNormal));
}

/*! This function provides an integrator for ideal light source, i.e. it evaluates 
   the reflection equation in the rather simple case where it is known that there is 
   incoming irradiance from exactly one direction. This boils down to multiplying 
   all inputs.
  \param OutgoingRadiance The outgoing radiance, which is the result of evaluating 
         the reflection equation.
  \param IrradianceCosineFactor See ComputeIrradianceCosineFactor().
  \param IdealIncomingIrradiance The incoming irradiance from the direction 
         IdealIncomingWorldSpaceLightDirection on a surface element orthogonal to 
		 this direction (i.e. with out the cosine term). Note that we actually have 
		 a non-zero irradiance from a single direction (i.e. an infinitely small 
		 solid angle) due to the use of ideal light sources. 
  \param IdealBRDFCoefficient The BRDF coefficient, which has been computed for the 
         direction IdealIncomingWorldSpaceLightDirection and the direction for which 
		 OutgoingRadiance is requested.*/
void IdealIntegrator(out float3 OutgoingRadiance,float IrradianceCosineFactor,float IdealBRDFCoefficient,float3 IdealIncomingIrradiance){
	OutgoingRadiance=(IrradianceCosineFactor*IdealBRDFCoefficient)*IdealIncomingIrradiance;
}


/*! This function provides an integrator for ambient light sources without 
   occlusion. It assumes that BRDFs are normalized to have a constant albedo of one 
   (before the actual albedo is multiplied onto that lateron). Thus, it really just 
   returns the appropriately scaled ambient radiance.
  \param OutgoingRadiance The outgoing radiance, which is the result of evaluating 
         the reflection equation.
  \param IncomingIrradiance The radiance that is coming in from every single 
		 direction.*/
void AmbientIntegrator(out float3 OutgoingRadiance,float3 IncomingRadiance){
	OutgoingRadiance=IncomingRadiance;
}
