/*! \file
   This file defines functions to define and render render participating media. Most 
   notably absorption and single scattering can be rendered.*/

#include "RandomNumbers.fx"
#include "ParticipatingMediaUtility.fx"

/*! This function prepares a matrix that is useful when working with non-linear 
   rectification transforms for directional lights and prefiltered single 
   scattering.
  \param OutRectificationToWorldSpace The rectification to world space transform. 
		 Rectification space is essentially light view space with the world space 
		 camera position at its origin. Additionally, the z and y axis are mirrored 
		 to ensure that inclinations grow as the depth grows. From rectification 
		 space coordinates rectified coordinates can be computed immediately.
  \param WorldSpaceCameraPosition The position of the camera for which single 
		 scattering is to be rendered in world space.
  \param ShadowMapViewToWorldSpace The view to world space transform used for the 
		 camera that renders the shadow map of the relevant directional light. It is 
		 assumed to be a rotation.*/
void ComputeRectificationToWorldSpaceDirectional(out float4x4 OutRectificationToWorldSpace,float3 WorldSpaceCameraPosition,float4x4 ShadowMapViewToWorldSpace){
	OutRectificationToWorldSpace=float4x4(
		 ShadowMapViewToWorldSpace._11, ShadowMapViewToWorldSpace._12, ShadowMapViewToWorldSpace._13, 0.0f,
		-ShadowMapViewToWorldSpace._21,-ShadowMapViewToWorldSpace._22,-ShadowMapViewToWorldSpace._23, 0.0f,
		-ShadowMapViewToWorldSpace._31,-ShadowMapViewToWorldSpace._32,-ShadowMapViewToWorldSpace._33, 0.0f,
		 WorldSpaceCameraPosition.x,    WorldSpaceCameraPosition.y,    WorldSpaceCameraPosition.z,    1.0f
	);
}


/*! This function computes the coordinate ranges covered by the rectified 
   coordinates of the specified view frustum.
  \param OutDistanceMinMax x is always set to zero, y is set to the maximal distance 
		 between the camera position and a point within the frustum.
  \param OutInclinationMinMax Minimal and maximal inclination of points within the 
		 frustum in rectification space. This ranges from 0 to pi.
  \param OutAzimuthMinMax Minimal and maximal azimuth of points within the frustum 
		 in rectification space. x and y-x are both always between 0 and 2*pi. The 
		 interval of radian angles inbetween covers the frustum.
  \param CameraProjectionToWorldSpace The projection to world space transform of the 
		 camera for which single scattering is to be rendered. This defines the 
		 frustum.
  \param CameraWorldToProjectionSpace Inverse of CameraProjectionToWorldSpace.
  \param RectificationToWorldSpace See 
		 ComputeRectificationToWorldSpaceDirectional().
  \param WorldToRectificationSpace Inverse of RectificationToWorldSpace.
  \note This is a reference implementation only. The function should be implemented 
		on the CPU side of your application and executed once per frame and light 
		source.*/
void GetRectifiedSpaceFrustumBounds(out float2 OutDistanceMinMax,out float2 OutInclinationMinMax,out float2 OutAzimuthMinMax,float4x4 CameraProjectionToWorldSpace,float4x4 CameraWorldToProjectionSpace,float4x4 RectificationToWorldSpace,float4x4 WorldToRectificationSpace){
	const float PI=3.1415926535897932384626433832795f;
	// Compute the rectification space corners of the far plane
	float4 pFarPlaneCorner[4]={
		float4(-1.0f,-1.0f,1.0f,1.0f),
		float4(-1.0f, 1.0f,1.0f,1.0f),
		float4( 1.0f, 1.0f,1.0f,1.0f),
		float4( 1.0f,-1.0f,1.0f,1.0f)
	};
	[unroll] for(int i=0;i!=4;++i){
		pFarPlaneCorner[i]=mul(pFarPlaneCorner[i],CameraProjectionToWorldSpace);
		pFarPlaneCorner[i]=mul(pFarPlaneCorner[i],WorldToRectificationSpace);
		pFarPlaneCorner[i]/=pFarPlaneCorner[i].w;
	}
	// Compute the maximal distance
	OutDistanceMinMax=float2(0.0f,0.0f);
	[unroll] for(int i=0;i!=4;++i){
		OutDistanceMinMax.y=max(OutDistanceMinMax.y,length(pFarPlaneCorner[i].xy));
	}
	// Figure out whether one of the poles of the unit sphere lies between the side 
	// clipping planes
	float4 NorthPole=mul(mul(float4(0.0f,0.0f,1.0f,1.0f),RectificationToWorldSpace),CameraWorldToProjectionSpace);
	bool4 ClipResult=bool4(NorthPole.x<=NorthPole.w,NorthPole.x>=-NorthPole.w,NorthPole.y<=NorthPole.w,NorthPole.y>=-NorthPole.w);
	bool NorthPoleInFrustum=all(ClipResult);
	bool SouthPoleInFrustum=all(!ClipResult);
	// Compute the azimuths
	if(NorthPoleInFrustum || SouthPoleInFrustum){
		OutAzimuthMinMax=float2(0.0f,2.0f*PI);
	}
	else{
		// Project down to the plane and normalize
		float2 pFarPlaneCornerFlat[4];
		[unroll] for(int i=0;i!=4;++i){
			pFarPlaneCornerFlat[i]=normalize(pFarPlaneCorner[i].xy);
		}
		// Compute the biggest pairwise angle
		float MinDotProduct=2.0f;
		float2 pCorner[2];
		[unroll] for(int i=0;i!=4;++i){
			[unroll] for(int j=0;j!=i;++j){
				float DotProduct=dot(pFarPlaneCornerFlat[i],pFarPlaneCornerFlat[j]);
				bool IsMinimal=(DotProduct<MinDotProduct);
				pCorner[0]=IsMinimal?pFarPlaneCornerFlat[i]:pCorner[0];
				pCorner[1]=IsMinimal?pFarPlaneCornerFlat[j]:pCorner[1];
				MinDotProduct=IsMinimal?DotProduct:MinDotProduct;
			}
		}
		// Figure out which of the two corners is clockwise of the other (when going 
		// no farther than pi)
		bool FirstIsClockwise=dot(float2(-pCorner[0].y,pCorner[0].x),pCorner[1])>0.0f;
		// Compute the angles
		float AzimuthRange=acos(MinDotProduct);
		float2 MinCorner=FirstIsClockwise?pCorner[0]:pCorner[1];
		OutAzimuthMinMax.x=atan2(MinCorner.y,MinCorner.x);
		OutAzimuthMinMax.x=(OutAzimuthMinMax.x<0.0f)?(OutAzimuthMinMax.x+2.0f*PI):OutAzimuthMinMax.x;
		OutAzimuthMinMax.y=OutAzimuthMinMax.x+AzimuthRange;
	}
	// Compute the inclinations. Start with the corner points.
	OutInclinationMinMax=float2(PI,0.0f);
	[unroll] for(int i=0;i!=4;++i){
		float CornerInclination=acos(normalize(pFarPlaneCorner[i].xyz).z);
		OutInclinationMinMax.x=min(OutInclinationMinMax.x,CornerInclination);
		OutInclinationMinMax.y=max(OutInclinationMinMax.y,CornerInclination);
	}
	// Now handle the case that one of the lines bounding the far plane holds an 
	// extremum
	[unroll] for(int i=0;i!=4;++i){
		float3 Start=pFarPlaneCorner[i].xyz;
		float3 End=pFarPlaneCorner[(i<3)?(i+1):0].xyz;
		float3 Offset=End-Start;
		// Determine the location of the extremal inclination on the line
		float InclinationExtremum=(Start.z*dot(Start,Offset)-Offset.z*dot(Start,Start))
								 /(Offset.z*dot(Start,Offset)-Start.z*dot(Offset,Offset));
		InclinationExtremum=saturate(InclinationExtremum);
		// Compute the inclination and update the minima and maxima
		float3 ExtremumLocation=normalize(lerp(Start,End,InclinationExtremum));
		float ExtremalLineInclination=acos(ExtremumLocation.z);
		OutInclinationMinMax.x=min(OutInclinationMinMax.x,ExtremalLineInclination);
		OutInclinationMinMax.y=max(OutInclinationMinMax.y,ExtremalLineInclination);
	}
	// Finally, handle the case that some point in the inner of the far plane is an 
	// extremum
	OutInclinationMinMax.x=NorthPoleInFrustum?0.0f:OutInclinationMinMax.x;
	OutInclinationMinMax.y=SouthPoleInFrustum?PI:OutInclinationMinMax.y;
	// Stretch inclination bounds a little to avoid unnecessary leaking
	OutInclinationMinMax=lerp(OutInclinationMinMax.xx,OutInclinationMinMax.yy,float2(-0.05f,1.05f));
	OutInclinationMinMax=clamp(OutInclinationMinMax,0.0f,PI);
}


/*! @{
   This function loads the samples of the given texture for the specified texel and 
   takes their average. This is a somewhat reasonable custom resolve for a depth 
   stencil texture in the context of single scattering because single scattering 
   depends monotonously (although not linearly) on depth.*/
void GetPerspectiveDepth1(out float OutPerspectiveDepth,Texture2DMS<float,1> DepthStencilTexture,int2 TexelIndex){
	uint nSample=1;
	OutPerspectiveDepth=0.0f;
	[unroll] for(uint i=0;i!=nSample;++i){
		OutPerspectiveDepth+=DepthStencilTexture.Load(TexelIndex,i);
	}
	OutPerspectiveDepth/=nSample;
}
void GetPerspectiveDepth2(out float OutPerspectiveDepth,Texture2DMS<float,2> DepthStencilTexture,int2 TexelIndex){
	uint nSample=2;
	OutPerspectiveDepth=0.0f;
	[unroll] for(uint i=0;i!=nSample;++i){
		OutPerspectiveDepth+=DepthStencilTexture.Load(TexelIndex,i);
	}
	OutPerspectiveDepth/=nSample;
}
void GetPerspectiveDepth4(out float OutPerspectiveDepth,Texture2DMS<float,4> DepthStencilTexture,int2 TexelIndex){
	uint nSample=4;
	OutPerspectiveDepth=0.0f;
	[unroll] for(uint i=0;i!=nSample;++i){
		OutPerspectiveDepth+=DepthStencilTexture.Load(TexelIndex,i);
	}
	OutPerspectiveDepth/=nSample;
}
void GetPerspectiveDepth8(out float OutPerspectiveDepth,Texture2DMS<float,8> DepthStencilTexture,int2 TexelIndex){
	uint nSample=8;
	OutPerspectiveDepth=0.0f;
	[unroll] for(uint i=0;i!=nSample;++i){
		OutPerspectiveDepth+=DepthStencilTexture.Load(TexelIndex,i);
	}
	OutPerspectiveDepth/=nSample;
}
//! @}


/*! This function computes the extinction coefficient of participating media from 
   the given half opacity distance.*/
void ComputeExtinctionCoefficient(out float OutExtinctionCoefficient,float HalfOpacityDistance){
	OutExtinctionCoefficient=-log(0.5f)/HalfOpacityDistance;
}


/*! This function computes the world space distance between the given camera 
   position and fragment position.*/
void ComputeViewRayLengthFromWorldSpace(out float OutViewRayLength,float3 WorldSpacePosition,float3 WorldSpaceCameraPosition){
	OutViewRayLength=distance(WorldSpacePosition,WorldSpaceCameraPosition);
}


/*! This function converts an outgoing radiance to an incoming radiance by taking 
   the attenuation due to absorption and outscattering in homogeneous participating 
   media into account.
  \param OutReceivedRadiance The radiance at the end of the considered ray.
  \param OutgoingRadiance The radiance at the beginning of the ray.
  \param ViewRayLength The length of the considered ray.
  \param ExtinctionCoefficient The extinction coefficient.*/
void ApplyAbsorptionToOutgoingRadiance(out float3 OutReceivedRadiance,float3 OutgoingRadiance,float ViewRayLength,float ExtinctionCoefficient){
	OutReceivedRadiance=OutgoingRadiance*exp(-ExtinctionCoefficient*ViewRayLength);
}


/*! This function computes the distance between a specified point in projection 
   space and the origin of the view space coordinate system (which is supposed to be 
   the camera location) in view space.
  \param OutViewRayLength The computed view space distance.
  \param OutWorldSpaceRayStart The world space position at which the ray starts 
		 (i.e. the camera position).
  \param OutWorldSpaceRayEnd The world space position at which the ray ends (i.e. 
		 the specified point).
  \param TexCoord The texture coordinate that corresponds to the point of interest 
		 in the render target.
  \param PerspectiveDepth The depth stored in the depth stencil texture for the 
		 point of interest. It is perspectivic (z/w) not linear (z).
  \param ProjectionToViewSpace The projection to view space transform of the 
		 camera.
  \param ViewToWorldSpace The view to world space transform of the camera.*/
void ComputeViewRayFromProjectionSpace(out float OutViewRayLength,out float3 OutWorldSpaceRayStart,out float3 OutWorldSpaceRayEnd,float2 TexCoord,float PerspectiveDepth,float4x4 ProjectionToViewSpace,float4x4 ViewToWorldSpace){
	float4 ProjectionSpacePosition=float4(TexCoord*float2(2.0f,-2.0f)+float2(-1.0f,1.0f),PerspectiveDepth,1.0f);
	float4 ViewSpacePosition=mul(ProjectionSpacePosition,ProjectionToViewSpace);
	ViewSpacePosition/=ViewSpacePosition.w;
	OutWorldSpaceRayStart=mul(float4(0.0f,0.0f,0.0f,1.0f),ViewToWorldSpace).xyz;
	OutWorldSpaceRayEnd=mul(ViewSpacePosition,ViewToWorldSpace).xyz;
	OutViewRayLength=length(ViewSpacePosition.xyz);
}


/*! This function computes the constant factors in front of the single scattering 
   integral divided by the extinction coefficient. This is simply the product of its 
   arguments.*/
void ComputeSingleScatteringFactors(out float3 OutPhaseTimesAlbedoTimesIrradiance,float3 PhaseTimesAlbedo,float3 DirectionalLightIrradiance){
	OutPhaseTimesAlbedoTimesIrradiance=PhaseTimesAlbedo*DirectionalLightIrradiance;
}

/*! Given the value of the single scattering integral times the extinction 
   coefficient this function computes the radiance received due to single 
   scattering. The value of the scattering integral has to be a value between zero 
   (no scattering) and one (maximal scattering).
  \sa ComputeSingleScatteringFactors() */
void ComputeSingleScatteringRadiance(out float3 OutScatteringRadiance,float ScatteringIntegral,float3 PhaseTimesAlbedoTimesIrradiance){
	OutScatteringRadiance=ScatteringIntegral*PhaseTimesAlbedoTimesIrradiance;
}

/*! Given the radiance value due to single scattering this function computes the 
   color to be output by the shader for additive blending. Essentially, it just sets 
   the alpha to one.
  \sa ComputeSingleScatteringRadiance() */
void ComputeScatteringColor(out float4 OutScatteringColor,float3 ScatteringRadiance){
	OutScatteringColor.rgb=ScatteringRadiance;
	OutScatteringColor.a=1.0f;
}


/*! This function evaluates the single scattering integral under the assumption that 
   the light is never occluded.
  \param OutScatteringIntegral See ComputeSingleScatteringRadiance().
  \param ViewRayLength The length of the view ray for which single scattering should 
		 be accumulated.
  \param ExtinctionCoefficient The extinction coefficient of the participating 
		 media.*/
void ComputeSingleScatteringNoOcclusionDirectional(out float OutScatteringIntegral,float ViewRayLength,float ExtinctionCoefficient){
	OutScatteringIntegral=1.0f-exp(-ExtinctionCoefficient*ViewRayLength);
}

/*! This function evaluates the scattering integral resolving occlusions by sampling 
   the given shadow map densely with some jittering to replace banding by white 
   noise.
  \param OutScatteringIntegral See ComputeSingleScatteringRadiance().
  \param ViewRayLength The length of the view ray for which single scattering should 
		 be accumulated.
  \param WorldSpaceViewRayStart,WorldSpaceViewRayEnd The endpoints of the view ray 
		 in world space coordinates.
  \param ShadowMapSampler A sampler for a shadow map of the scene.
  \param ShadowMapWorldToProjectionSpace The world to projection space transform of 
		 the provided shadow map.
  \param ExtinctionCoefficient The extinction coefficient of the participating 
		 media.
  \param nSample The number samples that is placed along the ray to evaluate the 
		 integral.*/
void ComputeSingleScatteringRayMarchingDirectional(out float OutScatteringIntegral,float ViewRayLength,float3 WorldSpaceViewRayStart,float3 WorldSpaceViewRayEnd,SSamplerTexturePair2D_float ShadowMapSampler,float4x4 ShadowMapWorldToProjectionSpace,float ExtinctionCoefficient,int nSample){
	OutScatteringIntegral=0.0f;
	// Convert ray start and ray end to projection space and compute the offset 
	// between consecutive samples
	float4 RayStart=mul(float4(WorldSpaceViewRayStart,1.0f),ShadowMapWorldToProjectionSpace);
	float4 RayEnd=mul(float4(WorldSpaceViewRayEnd,1.0f),ShadowMapWorldToProjectionSpace);
	// Convert the projection space coordinates to coordinates that enable immediate 
	// computation of texture coordinates to save two mads within the loop
	RayStart.xy=RayStart.xy*float2(0.5f,-0.5f)+RayStart.w*float2(0.5f,0.5f);
	RayEnd.xy=RayEnd.xy*float2(0.5f,-0.5f)+RayEnd.w*float2(0.5f,0.5f);
	// Implement jittering to avoid banding. Do this using a cheap linear 
	// congruential generator.
	float InvSampleCount=1.0f/nSample;
	float RandomOffset=GetRandomNumber(WorldSpaceViewRayEnd);
	float InitialLerpFactor=-RandomOffset*InvSampleCount;
	// Prepare variables that can be updated efficiently during iterations of the 
	// loop
	float4 SampleOffset=(RayEnd-RayStart)*InvSampleCount;
	float4 SamplePosition=lerp(RayStart,RayEnd,InitialLerpFactor);
	float Weight=exp(-ExtinctionCoefficient*InitialLerpFactor*ViewRayLength)-exp(-ExtinctionCoefficient*(InitialLerpFactor+InvSampleCount)*ViewRayLength);
	float WeightFactor=exp(-ExtinctionCoefficient*ViewRayLength*InvSampleCount);
	[loop] for(int i=0;i!=nSample;++i){
		// Compute the sample position in shadow map projection space
		SamplePosition+=SampleOffset;
		// Do the look up in the shadow map
		float InvSamplePositionW=1.0f/SamplePosition.w;
		float2 TexCoord=SamplePosition.xy*InvSamplePositionW;
		float RayDepth=SamplePosition.z*InvSamplePositionW;
		float ShadowMapDepth=ShadowMapSampler.Texture.SampleLevel(ShadowMapSampler.Sampler,TexCoord,0.0f);
		// Compute the contribution to the single scattering integral
		bool LightVisible=(RayDepth<ShadowMapDepth || TexCoord.x<0.0f || TexCoord.y<0.0f || TexCoord.x>1.0f || TexCoord.y>1.0f);
		Weight*=WeightFactor;
		OutScatteringIntegral+=LightVisible?Weight:0.0f;
	}
}


/*! This function computes coordinates which are needed for evaluation of 
   prefiltered single scattering with any prefiltered, non-linearly rectified shadow 
   map.
  \param OutRayEndTexCoord The texture coordinate at which the prefiltered shadow 
		 map has to be sampled to get information about the provided ray.
  \param OutRayDepth A size that depends linearly on the inclination of the relevant 
         ray. It functions as shadow map depth value in the rectified coordinate 
		 system. -1 corresponds to InclinationMinMax.x, 1 corresponds to 
		 InclinationMinMax.y.
  \param WorldSpaceViewRayEnd The endpoint of the view ray in world space 
		 coordinates.
  \param WorldToRectificationSpace Inverse of the transform provided by 
		 ComputeRectificationToWorldSpaceDirectional().
  \param DistanceMinMax,InclinationMinMax,AzimuthMinMax See 
		 GetRectifiedSpaceFrustumBounds().
  \sa ApplyTransmittanceWeightedPrefixSumNonLinearRectification() */
void ComputePrefilteredSingleScatteringCoordinatesNonLinearRectification(out float2 OutRayEndTexCoord,out float OutRayDepth,float3 WorldSpaceViewRayEnd,float4x4 WorldToRectificationSpace,float2 DistanceMinMax,float2 InclinationMinMax,float2 AzimuthMinMax){
	const float PI=3.1415926535897932384626433832795f;
	// Compute rectified coordinates for the view ray end in rectification space
	float4 RectificationSpaceViewRayEnd=mul(float4(WorldSpaceViewRayEnd,1.0f),WorldToRectificationSpace);
	float Distance,Inclination,Azimuth;
	CartesianToRectified(Distance,Inclination,Azimuth,RectificationSpaceViewRayEnd.xyz);
	// Normalize the coordinates
	float InvMaxDistance=1.0f/DistanceMinMax.y;
	OutRayEndTexCoord.x=Distance*InvMaxDistance;
	OutRayEndTexCoord.y=Azimuth-AzimuthMinMax.x;
	OutRayEndTexCoord.y=(OutRayEndTexCoord.y<0.0f)?(OutRayEndTexCoord.y+2.0f*PI):OutRayEndTexCoord.y;
	float InvAzimuthRange=1.0f/(AzimuthMinMax.y-AzimuthMinMax.x);
	OutRayEndTexCoord.y*=InvAzimuthRange;
	OutRayDepth=mad((Inclination-InclinationMinMax.x)/(InclinationMinMax.y-InclinationMinMax.x),2.0f,-1.0f);
}


/*! This function evaluates the scattering integral resolving occlusions by means of 
   a rectified moment shadow map holding weighted prefix sums of rows.
  \param OutScatteringIntegral See ComputeSingleScatteringRadiance().
  \param RayEndTexCoord,RayDepth See 
		 ComputePrefilteredSingleScatteringCoordinatesNonLinearRectification().
  \param ScatteringIntegralWithoutOcclusion The output value in absence of shadow.
  \param ShadowMapSampler0,ShadowMapSampler1,ShadowMapSampler2,ShadowMapSampler3 
		 Samplers for the rectified and prefiltered shadow maps holding  of the 
		 scene.*/
void ComputePrefilteredSingleScatteringConvolution8(out float OutScatteringIntegral,float2 RayEndTexCoord,float RayDepth,float ScatteringIntegralWithoutOcclusion,SSamplerTexturePair2D ShadowMapSampler0,SSamplerTexturePair2D ShadowMapSampler1,SSamplerTexturePair2D ShadowMapSampler2,SSamplerTexturePair2D ShadowMapSampler3){
	const float PI=3.1415926535897932384626433832795f;
	// Get the Fourier coefficients from the textures
	float4 Fourier1_2=ShadowMapSampler0.Texture.Sample(ShadowMapSampler0.Sampler,RayEndTexCoord);
	float4 Fourier3_4=ShadowMapSampler1.Texture.Sample(ShadowMapSampler1.Sampler,RayEndTexCoord);
	float4 Fourier5_6=ShadowMapSampler2.Texture.Sample(ShadowMapSampler2.Sampler,RayEndTexCoord);
	float4 Fourier7_8=ShadowMapSampler3.Texture.Sample(ShadowMapSampler3.Sampler,RayEndTexCoord);
	float2 pFourierCoefficient[8]={
		Fourier1_2.xy*2.0f-1.0f,
		Fourier1_2.zw*2.0f-1.0f,
		Fourier3_4.zw*2.0f-1.0f,
		Fourier3_4.zw*2.0f-1.0f,
		Fourier5_6.zw*2.0f-1.0f,
		Fourier5_6.zw*2.0f-1.0f,
		Fourier7_8.zw*2.0f-1.0f,
		Fourier7_8.zw*2.0f-1.0f
	};
	// Evaluate the first complex basis function
	float2 Fourier1;
	sincos(mad(RayDepth,0.5f*PI,0.5f*PI),Fourier1.y,Fourier1.x);
	// Evaluate the other basis functions using complex exponentiation
	float2 pBasisFunction[8];
	pBasisFunction[0]=Fourier1;
	float2 FourierSquare=float2(dot(Fourier1,float2(Fourier1.x,-Fourier1.y)),dot(Fourier1.xy,Fourier1.yx));
	[unroll] for(int i=1;i!=8;++i){
		float2 LHS=pBasisFunction[i-1];
		float2 RHS=FourierSquare;
		pBasisFunction[i]=float2(dot(LHS,float2(RHS.x,-RHS.y)),dot(LHS.xy,RHS.yx));
	}
	// Compute the shadow intensity which is a linear combination of the basis 
	// functions using the Fourier coefficients as weights
	OutScatteringIntegral=0.5f;
	[unroll] for(int i=0;i!=8;++i){
		float Constant=PI*(2.0f*i+1.0f);
		float Summand=pBasisFunction[i].y*pFourierCoefficient[i].x-pBasisFunction[i].x*pFourierCoefficient[i].y;
		OutScatteringIntegral+=Summand*2.0f/Constant;
	}
	OutScatteringIntegral=saturate(1.0f-OutScatteringIntegral);
	OutScatteringIntegral*=ScatteringIntegralWithoutOcclusion;
}


/*! Like ComputePrefilteredSingleScatteringConvolution16() but for eight Fourier 
   coefficients.*/
void ComputePrefilteredSingleScatteringConvolution16(out float OutScatteringIntegral,float2 RayEndTexCoord,float RayDepth,float ScatteringIntegralWithoutOcclusion,SSamplerTexturePair2D ShadowMapSampler0,SSamplerTexturePair2D ShadowMapSampler1,SSamplerTexturePair2D ShadowMapSampler2,SSamplerTexturePair2D ShadowMapSampler3,SSamplerTexturePair2D ShadowMapSampler4,SSamplerTexturePair2D ShadowMapSampler5,SSamplerTexturePair2D ShadowMapSampler6,SSamplerTexturePair2D ShadowMapSampler7){
	const float PI=3.1415926535897932384626433832795f;
	// Get the Fourier coefficients from the textures
	float4 Fourier1_2=ShadowMapSampler0.Texture.Sample(ShadowMapSampler0.Sampler,RayEndTexCoord);
	float4 Fourier3_4=ShadowMapSampler1.Texture.Sample(ShadowMapSampler1.Sampler,RayEndTexCoord);
	float4 Fourier5_6=ShadowMapSampler2.Texture.Sample(ShadowMapSampler2.Sampler,RayEndTexCoord);
	float4 Fourier7_8=ShadowMapSampler3.Texture.Sample(ShadowMapSampler3.Sampler,RayEndTexCoord);
	float4 Fourier9_10=ShadowMapSampler4.Texture.Sample(ShadowMapSampler4.Sampler,RayEndTexCoord);
	float4 Fourier11_12=ShadowMapSampler5.Texture.Sample(ShadowMapSampler5.Sampler,RayEndTexCoord);
	float4 Fourier13_14=ShadowMapSampler6.Texture.Sample(ShadowMapSampler6.Sampler,RayEndTexCoord);
	float4 Fourier15_16=ShadowMapSampler7.Texture.Sample(ShadowMapSampler7.Sampler,RayEndTexCoord);
	float2 pFourierCoefficient[16]={
		Fourier1_2.xy*2.0f-1.0f,
		Fourier1_2.zw*2.0f-1.0f,
		Fourier3_4.zw*2.0f-1.0f,
		Fourier3_4.zw*2.0f-1.0f,
		Fourier5_6.zw*2.0f-1.0f,
		Fourier5_6.zw*2.0f-1.0f,
		Fourier7_8.zw*2.0f-1.0f,
		Fourier7_8.zw*2.0f-1.0f,
		Fourier9_10.xy*2.0f-1.0f,
		Fourier9_10.zw*2.0f-1.0f,
		Fourier11_12.xy*2.0f-1.0f,
		Fourier11_12.zw*2.0f-1.0f,
		Fourier13_14.xy*2.0f-1.0f,
		Fourier13_14.zw*2.0f-1.0f,
		Fourier15_16.xy*2.0f-1.0f,
		Fourier15_16.zw*2.0f-1.0f
	};
	// Evaluate the first complex basis function
	float2 Fourier1;
	sincos(mad(RayDepth,0.5f*PI,0.5f*PI),Fourier1.y,Fourier1.x);
	// Evaluate the other basis functions using complex exponentiation
	float2 pBasisFunction[16];
	pBasisFunction[0]=Fourier1;
	float2 FourierSquare=float2(dot(Fourier1,float2(Fourier1.x,-Fourier1.y)),dot(Fourier1.xy,Fourier1.yx));
	[unroll] for(int i=1;i!=16;++i){
		float2 LHS=pBasisFunction[i-1];
		float2 RHS=FourierSquare;
		pBasisFunction[i]=float2(dot(LHS,float2(RHS.x,-RHS.y)),dot(LHS.xy,RHS.yx));
	}
	// Compute the shadow intensity which is a linear combination of the basis 
	// functions using the Fourier coefficients as weights
	OutScatteringIntegral=0.5f;
	[unroll] for(int i=0;i!=16;++i){
		float Constant=PI*(2.0f*i+1.0f);
		float Summand=pBasisFunction[i].y*pFourierCoefficient[i].x-pBasisFunction[i].x*pFourierCoefficient[i].y;
		OutScatteringIntegral+=Summand*2.0f/Constant;
	}
	OutScatteringIntegral=saturate(1.0f-OutScatteringIntegral);
	OutScatteringIntegral*=ScatteringIntegralWithoutOcclusion;
}


/*! This function computes an adaptive overestimation weight to avoid common 
   artifacts at epipoles. It works for directional light sources only. When looking 
   directly into the light minimal overestimation is used, when looking directly 
   away from the light maximal overestimation is used (on a per pixel basis). In 
   between a continuous and monotonous function of inclination is used.
  \param OutOverestimationWeight Zero if single scattering is to be underestimated. 
		 One if single scattering is to be overestimated. In between radiance is 
		 interpolated linearly.
  \param WorldSpaceViewRayStart,WorldSpaceViewRayEnd The endpoints of the view ray 
		 in world space coordinates.
  \param WorldSpaceLightDirection The normalized world space light direction of the 
		 directional light.
  \param OverestimationWeightMinMax The minimal possible overestimation weight in x, 
		 the maximal weight in y.*/
void ComputeAdaptiveOverestimationWeight(out float OutOverestimationWeight,float3 WorldSpaceViewRayStart,float3 WorldSpaceViewRayEnd,float3 WorldSpaceLightDirection,float2 OverestimationWeightMinMax){
	// Compute the z-coordinate of the normalized view ray direction in an 
	// orthonormal coordinate system where the light direction corresponds to the 
	// positive z-axis
	float3 WorldSpaceViewRayDirection=normalize(WorldSpaceViewRayEnd-WorldSpaceViewRayStart);
	float LightViewZ=dot(WorldSpaceLightDirection,WorldSpaceViewRayDirection);
	// Interpolate linearly between minimal and maximal overestimation weight using 
	// this z-coordinate
	float LerpFactor=mad(LightViewZ,0.5f,0.5f);
	OutOverestimationWeight=saturate(lerp(OverestimationWeightMinMax.x,OverestimationWeightMinMax.y,LerpFactor));
}


/*! This function evaluates the scattering integral resolving occlusions by means of 
   a rectified moment shadow map holding weighted prefix sums of rows.
  \param OutScatteringIntegral See ComputeSingleScatteringRadiance().
  \param RayEndTexCoord,RayDepth See 
		 ComputePrefilteredSingleScatteringCoordinatesNonLinearRectification().
  \param ScatteringIntegralWithoutOcclusion The output value in absence of 
		 occlusion.
  \param ShadowMapSampler0 A sampler for the rectified and prefiltered moment shadow 
		 map of the scene.
  \param MomentBias The bias used to overcome quantization artifacts in the moment 
		 shadow map.
  \param OverestimationWeight If this value is zero, the output underestimates the 
		 actual amount of single scattering. If it is one, it overestimates. 
		 Inbetween linear interpolation is employed.*/
void ComputePrefilteredSingleScattering4Moments(out float OutScatteringIntegral,float2 RayEndTexCoord,float RayDepth,float ScatteringIntegralWithoutOcclusion,SSamplerTexturePair2D ShadowMapSampler0,float MomentBias,float OverestimationWeight=0.5f){
	float4 OptimizedSample=ShadowMapSampler0.Texture.Sample(ShadowMapSampler0.Sampler,RayEndTexCoord);
	OutScatteringIntegral=1.0f-EstimateIntegralFrom4Moments(RayDepth,OptimizedSample,MomentBias,1.0f-OverestimationWeight);
	OutScatteringIntegral*=ScatteringIntegralWithoutOcclusion;
}


/*! Like ComputePrefilteredSingleScattering4Moments() but uses six moments.*/
void ComputePrefilteredSingleScattering2_4Moments(out float OutScatteringIntegral,float2 RayEndTexCoord,float RayDepth,float ScatteringIntegralWithoutOcclusion,SSamplerTexturePair2D_float2 ShadowMapSampler0,SSamplerTexturePair2D ShadowMapSampler1,float MomentBias,float OverestimationWeight=0.5f){
	float2 OptimizedSample0=ShadowMapSampler0.Texture.Sample(ShadowMapSampler0.Sampler,RayEndTexCoord);
	float4 OptimizedSample1=ShadowMapSampler1.Texture.Sample(ShadowMapSampler1.Sampler,RayEndTexCoord);
	OutScatteringIntegral=1.0f-EstimateIntegralFrom6Moments(RayDepth,OptimizedSample0,OptimizedSample1,MomentBias,1.0f-OverestimationWeight);
	OutScatteringIntegral*=ScatteringIntegralWithoutOcclusion;
}


/*! Like ComputePrefilteredSingleScattering4Moments() but uses six moments.*/
void ComputePrefilteredSingleScattering3_3Moments(out float OutScatteringIntegral,float2 RayEndTexCoord,float RayDepth,float ScatteringIntegralWithoutOcclusion,SSamplerTexturePair2D_float3 ShadowMapSampler0,SSamplerTexturePair2D_float3 ShadowMapSampler1,float MomentBias,float OverestimationWeight=0.5f){
	float3 OptimizedSample0=ShadowMapSampler0.Texture.Sample(ShadowMapSampler0.Sampler,RayEndTexCoord);
	float3 OptimizedSample1=ShadowMapSampler1.Texture.Sample(ShadowMapSampler1.Sampler,RayEndTexCoord);
	OutScatteringIntegral=1.0f-EstimateIntegralFrom6Moments(RayDepth,OptimizedSample0.xy,float4(OptimizedSample0.z,OptimizedSample1),MomentBias,1.0f-OverestimationWeight);
	OutScatteringIntegral*=ScatteringIntegralWithoutOcclusion;
}


/*! This function computes the number of thread groups to be dispatched for 
   generation of row-wise prefix sums.
  \param nOutThreadGroup The number of thread groups along each dimension to be 
		 dispatched.
  \param ThreadGroupSize The size of thread groups along each dimension for the 
		 prefix sum generation. Only the second entry is supposed to be not one.
  \param nPrefixSumShadowMap The number of textures for which prefix sums need to be 
		 created. It is multiplied onto the number of vertical thread groups.
  \param TextureSize The width and height of the texture for which prefix sums 
		 should be generated in x and y.
  \note This function is provided for reference only and would typically be 
		implemented on the CPU-side of the application.*/
void GetPrefixSumThreadGroupCount(out uint3 nOutThreadGroup,uint3 ThreadGroupSize,uint nPrefixSumShadowMap,float4 TextureSize){
	uint Height=(uint)(TextureSize.y);
	nOutThreadGroup=uint3(1,nPrefixSumShadowMap*Height/ThreadGroupSize.y,1);
}


/*! Constructs a tuple (Gradient,YIntercept) describing the linear transform that 
   maps depths in shadow map space to z values in rectification space.*/
void GetShadowMapToRectificationDepthTransform(out float2 OutShadowMapToRectificationDepthTransform,float4x4 RectificationToShadowMapProjectionSpace){
	float4 ProjectionSpaceCameraPosition=mul(float4(0.0f,0.0f,0.0f,1.0f),RectificationToShadowMapProjectionSpace);
	float CameraShadowMapDepth=ProjectionSpaceCameraPosition.z/ProjectionSpaceCameraPosition.w;
	float4 ProjectionSpaceNorthPole=mul(float4(0.0f,0.0f,1.0f,1.0f),RectificationToShadowMapProjectionSpace);
	float NorthPoleShadowMapDepth=ProjectionSpaceNorthPole.z/ProjectionSpaceNorthPole.w;
	OutShadowMapToRectificationDepthTransform.x=1.0f/(NorthPoleShadowMapDepth-CameraShadowMapDepth);
	OutShadowMapToRectificationDepthTransform.y=-CameraShadowMapDepth*OutShadowMapToRectificationDepthTransform.x;
	OutShadowMapToRectificationDepthTransform.x*=0.5f;
	OutShadowMapToRectificationDepthTransform.y+=OutShadowMapToRectificationDepthTransform.x;
}


/*! This function takes a single sample from the given non-rectified shadow map at 
   the location that corresponds to the given rectified texture coordinate. It then 
   converts the sampled depth to an inclination and maps it to the appropriate range 
   to get a depth in the rectified coordinate system and outputs it.*/
void GetRectifiedDepth(out float OutRectifiedDepth,SSamplerTexturePair2D_float ShadowMapSampler0,float2 RectifiedTexCoord,float4x4 RectificationToShadowMapProjectionSpace,float2 ShadowMapToRectificationDepthTransform,float2 DistanceMinMax,float2 InclinationMinMax,float2 AzimuthMinMax){
	// Convert the rectified texture coordinate to a non-rectified one
	float Azimuth=lerp(AzimuthMinMax.x,AzimuthMinMax.y,RectifiedTexCoord.y);
	float Distance=RectifiedTexCoord.x*DistanceMinMax.y;
	float2 RectificationSpaceXY;
	sincos(Azimuth,RectificationSpaceXY.y,RectificationSpaceXY.x);
	RectificationSpaceXY*=Distance;
	float4 ShadowMapProjectionSpace=mul(float4(RectificationSpaceXY,0.0f,1.0f),RectificationToShadowMapProjectionSpace);
	ShadowMapProjectionSpace/=ShadowMapProjectionSpace.w;
	float2 ShadowMapTexCoord=mad(ShadowMapProjectionSpace.xy,float2(0.5f,-0.5f),float2(0.5f,0.5f));
	// Sample the shadow map
	float ShadowMapDepth=mad(ShadowMapSampler0.Texture.Sample(ShadowMapSampler0.Sampler,ShadowMapTexCoord),2.0f,-1.0f);
	// Convert the linear depth to an inclination
	float RectificationSpaceDepth=mad(ShadowMapDepth,ShadowMapToRectificationDepthTransform.x,ShadowMapToRectificationDepthTransform.y);
	float Inclination=atan2(Distance,RectificationSpaceDepth);
	// Now convert to a rectified depth
	OutRectifiedDepth=saturate((Inclination-InclinationMinMax.x)/(InclinationMinMax.y-InclinationMinMax.x));
	OutRectifiedDepth=OutRectifiedDepth*2.0f-1.0f;
}


/*! This function takes a single filtered sample from the given non-rectified, 
   optimized four moment shadow map at the location that corresponds to the given 
   rectified texture coordinate. It then represents the first three moments by a 
   distribution using two depth values. The depth values are converted to 
   inclinations and mapped to the appropriate range to get a depth in the rectified 
   coordinate system. The two depth values and the weight for the second depth are 
   output.*/
void GetSparseRectifiedRepresentation(out float2 OutRectifiedDepth,out float OutDepthWeight1,SSamplerTexturePair2D ShadowMapSampler0,float2 RectifiedTexCoord,float4x4 RectificationToShadowMapProjectionSpace,float2 ShadowMapToRectificationDepthTransform,float2 DistanceMinMax,float2 InclinationMinMax,float2 AzimuthMinMax){
	// Convert the rectified texture coordinate to a non-rectified one
	float Azimuth=lerp(AzimuthMinMax.x,AzimuthMinMax.y,RectifiedTexCoord.y);
	float Distance=RectifiedTexCoord.x*DistanceMinMax.y;
	float2 RectificationSpaceXY;
	sincos(Azimuth,RectificationSpaceXY.y,RectificationSpaceXY.x);
	RectificationSpaceXY*=Distance;
	float4 ShadowMapProjectionSpace=mul(float4(RectificationSpaceXY,0.0f,1.0f),RectificationToShadowMapProjectionSpace);
	ShadowMapProjectionSpace/=ShadowMapProjectionSpace.w;
	float2 ShadowMapTexCoord=mad(ShadowMapProjectionSpace.xy,float2(0.5f,-0.5f),float2(0.5f,0.5f));
	// Sample the moment shadow map and represent its first three moments by two 
	// depth
	float4 OptimizedMoments=ShadowMapSampler0.Texture.Sample(ShadowMapSampler0.Sampler,ShadowMapTexCoord);
	float2 ShadowMapDepth;
	GetSparseRepresentation4Moments(ShadowMapDepth,OutDepthWeight1,OptimizedMoments);
	// Convert the linear depths to inclinations
	float2 RectificationSpaceDepth=mad(ShadowMapDepth,ShadowMapToRectificationDepthTransform.x,ShadowMapToRectificationDepthTransform.y);
	float2 Inclination=float2(
		atan2(Distance,RectificationSpaceDepth[0]),
		atan2(Distance,RectificationSpaceDepth[1])
	);
	// Now convert them to rectified depths
	OutRectifiedDepth=saturate((Inclination-InclinationMinMax.x)/(InclinationMinMax.y-InclinationMinMax.x));
	OutRectifiedDepth=OutRectifiedDepth*2.0f-1.0f;
}


/*! Converts a depth to four optimized moments.*/
void GetMomentsFromDepth4(out float4 OutQuantizedMoments0,float RectifiedDepth){
	ComputeMomentVector4MomentsOptimized(OutQuantizedMoments0,RectifiedDepth);
}

/*! Converts a distribution with two depth values to four optimized moments.
  \sa GetSparseRectifiedRepresentation() */
void GetMomentsFromSparseRectifiedRepresentation4(out float4 OutQuantizedMoments0,float2 RectifiedDepth,float DepthWeight1){
	float4 pCanonicalMoments0[2];
	[unroll] for(int i=0;i!=2;++i){
		float Depth=RectifiedDepth[i];
		float SquaredDepth=Depth*Depth;
		pCanonicalMoments0[i]=float4(Depth,SquaredDepth,SquaredDepth*Depth,SquaredDepth*SquaredDepth);
	}
	float4 QuantizedMoments0;
	Convert4MomentCanonicalToOptimized(QuantizedMoments0,lerp(pCanonicalMoments0[0],pCanonicalMoments0[1],DepthWeight1));
	OutQuantizedMoments0=QuantizedMoments0;
}


/*! Converts a depth to six optimized moments.*/
void GetMomentsFromDepth3_3(out float3 OutQuantizedMoments0,out float3 OutQuantizedMoments1,float RectifiedDepth){
	float2 QuantizedMoments0;
	float4 QuantizedMoments1;
	ComputeMomentVector6MomentsOptimized(QuantizedMoments0,QuantizedMoments1,RectifiedDepth);
	OutQuantizedMoments0=float3(QuantizedMoments0,QuantizedMoments1.x);
	OutQuantizedMoments1=QuantizedMoments1.yzw;
}

/*! Converts a distribution with two depth values to six optimized moments, output 
   as two three-element vectors.
  \sa GetSparseRectifiedRepresentation() */
void GetMomentsFromSparseRectifiedRepresentation3_3(out float3 OutQuantizedMoments0,out float3 OutQuantizedMoments1,float2 RectifiedDepth,float DepthWeight1){
	float2 pCanonicalMoments0[2];
	float4 pCanonicalMoments1[2];
	[unroll] for(int i=0;i!=2;++i){
		float Depth=RectifiedDepth[i];
		float SquaredDepth=Depth*Depth;
		float CubedDepth=SquaredDepth*Depth;
		pCanonicalMoments0[i]=float2(Depth,SquaredDepth);
		pCanonicalMoments1[i]=float4(CubedDepth,SquaredDepth*SquaredDepth,CubedDepth*SquaredDepth,CubedDepth*CubedDepth);
	}
	float2 QuantizedMoments0;
	float4 QuantizedMoments1;
	Convert6MomentCanonicalToOptimized(QuantizedMoments0,QuantizedMoments1,
		lerp(pCanonicalMoments0[0],pCanonicalMoments0[1],DepthWeight1),lerp(pCanonicalMoments1[0],pCanonicalMoments1[1],DepthWeight1));
	OutQuantizedMoments0.xy=QuantizedMoments0;
	OutQuantizedMoments0.z=QuantizedMoments1.x;
	OutQuantizedMoments1=QuantizedMoments1.yzw;
}

/*! Converts a depth to six optimized moments.*/
void GetMomentsFromDepth2_4(out float2 OutQuantizedMoments0,out float4 OutQuantizedMoments1,float RectifiedDepth){
	ComputeMomentVector6MomentsOptimized(OutQuantizedMoments0,OutQuantizedMoments1,RectifiedDepth);
}

/*! Converts a distribution with two depth values to six optimized moments, output 
   as one two-dimensional vector and one four-dimensional vector.
  \sa GetSparseRectifiedRepresentation() */
void GetMomentsFromSparseRectifiedRepresentation2_4(out float2 OutQuantizedMoments0,out float4 OutQuantizedMoments1,float2 RectifiedDepth,float DepthWeight1){
	float2 pCanonicalMoments0[2];
	float4 pCanonicalMoments1[2];
	[unroll] for(int i=0;i!=2;++i){
		float Depth=RectifiedDepth[i];
		float SquaredDepth=Depth*Depth;
		float CubedDepth=SquaredDepth*Depth;
		pCanonicalMoments0[i]=float2(Depth,SquaredDepth);
		pCanonicalMoments1[i]=float4(CubedDepth,SquaredDepth*SquaredDepth,CubedDepth*SquaredDepth,CubedDepth*CubedDepth);
	}
	Convert6MomentCanonicalToOptimized(OutQuantizedMoments0,OutQuantizedMoments1,
		lerp(pCanonicalMoments0[0],pCanonicalMoments0[1],DepthWeight1),lerp(pCanonicalMoments1[0],pCanonicalMoments1[1],DepthWeight1));
}


/*! Like ApplyTransmittanceWeightedPrefixSumNonLinearRectificationMSM4() but 
   operates on a convolution shadow map with eight complex Fourier coefficients.*/
void ApplyTransmittanceWeightedPrefixSumNonLinearRectificationCSM8(inout RWTexture2D<float4> OutFilteredShadowMap0,inout RWTexture2D<float4> OutFilteredShadowMap1,inout RWTexture2D<float4> OutFilteredShadowMap2,inout RWTexture2D<float4> OutFilteredShadowMap3,SSamplerTexturePair2D ShadowMapSampler0,SSamplerTexturePair2D ShadowMapSampler1,SSamplerTexturePair2D ShadowMapSampler2,SSamplerTexturePair2D ShadowMapSampler3,uint3 ThreadID,float ExtinctionCoefficient,float2 DistanceMinMax,float2 InclinationMinMax){
	uint Width,Height;
	OutFilteredShadowMap0.GetDimensions(Width,Height);
	float TexelTransmittance=ComputeTexelTransmittance(Width,ExtinctionCoefficient,DistanceMinMax,InclinationMinMax);
	uint iShadowMap=ThreadID.y/Height;
	uint iThreadY=ThreadID.y%Height;
	if(iShadowMap==0){
		ApplyTransmittanceWeightedPrefixSum(OutFilteredShadowMap0,ShadowMapSampler0.Texture,ThreadID.x,iThreadY,TexelTransmittance);
	}
	else if(iShadowMap==1){
		ApplyTransmittanceWeightedPrefixSum(OutFilteredShadowMap1,ShadowMapSampler1.Texture,ThreadID.x,iThreadY,TexelTransmittance);
	}
	else if(iShadowMap==2){
		ApplyTransmittanceWeightedPrefixSum(OutFilteredShadowMap2,ShadowMapSampler2.Texture,ThreadID.x,iThreadY,TexelTransmittance);
	}
	else if(iShadowMap==3){
		ApplyTransmittanceWeightedPrefixSum(OutFilteredShadowMap3,ShadowMapSampler3.Texture,ThreadID.x,iThreadY,TexelTransmittance);
	}
}


/*! Like ApplyTransmittanceWeightedPrefixSumNonLinearRectificationMSM4() but 
   operates on a convolution shadow map with 16 complex Fourier coefficients.*/
void ApplyTransmittanceWeightedPrefixSumNonLinearRectificationCSM16(inout RWTexture2D<float4> OutFilteredShadowMap0,inout RWTexture2D<float4> OutFilteredShadowMap1,inout RWTexture2D<float4> OutFilteredShadowMap2,inout RWTexture2D<float4> OutFilteredShadowMap3,inout RWTexture2D<float4> OutFilteredShadowMap4,inout RWTexture2D<float4> OutFilteredShadowMap5,inout RWTexture2D<float4> OutFilteredShadowMap6,inout RWTexture2D<float4> OutFilteredShadowMap7,SSamplerTexturePair2D ShadowMapSampler0,SSamplerTexturePair2D ShadowMapSampler1,SSamplerTexturePair2D ShadowMapSampler2,SSamplerTexturePair2D ShadowMapSampler3,SSamplerTexturePair2D ShadowMapSampler4,SSamplerTexturePair2D ShadowMapSampler5,SSamplerTexturePair2D ShadowMapSampler6,SSamplerTexturePair2D ShadowMapSampler7,uint3 ThreadID,float ExtinctionCoefficient,float2 DistanceMinMax,float2 InclinationMinMax){
	uint Width,Height;
	OutFilteredShadowMap0.GetDimensions(Width,Height);
	float TexelTransmittance=ComputeTexelTransmittance(Width,ExtinctionCoefficient,DistanceMinMax,InclinationMinMax);
	uint iShadowMap=ThreadID.y/Height;
	uint iThreadY=ThreadID.y%Height;
	if(iShadowMap==0){
		ApplyTransmittanceWeightedPrefixSum(OutFilteredShadowMap0,ShadowMapSampler0.Texture,ThreadID.x,iThreadY,TexelTransmittance);
	}
	else if(iShadowMap==1){
		ApplyTransmittanceWeightedPrefixSum(OutFilteredShadowMap1,ShadowMapSampler1.Texture,ThreadID.x,iThreadY,TexelTransmittance);
	}
	else if(iShadowMap==2){
		ApplyTransmittanceWeightedPrefixSum(OutFilteredShadowMap2,ShadowMapSampler2.Texture,ThreadID.x,iThreadY,TexelTransmittance);
	}
	else if(iShadowMap==3){
		ApplyTransmittanceWeightedPrefixSum(OutFilteredShadowMap3,ShadowMapSampler3.Texture,ThreadID.x,iThreadY,TexelTransmittance);
	}
	else if(iShadowMap==4){
		ApplyTransmittanceWeightedPrefixSum(OutFilteredShadowMap4,ShadowMapSampler4.Texture,ThreadID.x,iThreadY,TexelTransmittance);
	}
	else if(iShadowMap==5){
		ApplyTransmittanceWeightedPrefixSum(OutFilteredShadowMap5,ShadowMapSampler5.Texture,ThreadID.x,iThreadY,TexelTransmittance);
	}
	else if(iShadowMap==6){
		ApplyTransmittanceWeightedPrefixSum(OutFilteredShadowMap6,ShadowMapSampler6.Texture,ThreadID.x,iThreadY,TexelTransmittance);
	}
	else if(iShadowMap==7){
		ApplyTransmittanceWeightedPrefixSum(OutFilteredShadowMap7,ShadowMapSampler7.Texture,ThreadID.x,iThreadY,TexelTransmittance);
	}
}


/*! Generates transmittance weighted prefix sums for the given four moment shadow 
   map. The moment shadow map is supposed to be rectified, i.e. the texture 
   coordinates and depths refer to the coordinate system produced by 
   ComputePrefilteredSingleScatteringCoordinatesNonLinearRectification(). 
   Transmittance weighting means that summands are weighted by their (approximate) 
   contribution to the single scattering integral.
  \param OutFilteredShadowMap0 The texture which is to be overwritten by weighted 
		 prefix sums of moments.
  \param ShadowMapSampler0 A texture and sampler providing access to a rectified 
		 four moment shadow map for which prefix sums should be generated.
  \param ThreadID The y component of this vector indicates the row on which to 
		 operate.
  \param ExtinctionCoefficient The extinction coefficient of the participating 
		 media.
  \param DistanceMinMax,InclinationMinMax See GetRectifiedSpaceFrustumBounds().*/
void ApplyTransmittanceWeightedPrefixSumNonLinearRectificationMSM4(inout RWTexture2D<float4> OutFilteredShadowMap0,SSamplerTexturePair2D ShadowMapSampler0,uint3 ThreadID,float ExtinctionCoefficient,float2 DistanceMinMax,float2 InclinationMinMax){
	uint Width,Height;
	OutFilteredShadowMap0.GetDimensions(Width,Height);
	float TexelTransmittance=ComputeTexelTransmittance(Width,ExtinctionCoefficient,DistanceMinMax,InclinationMinMax);
	ApplyTransmittanceWeightedPrefixSum(OutFilteredShadowMap0,ShadowMapSampler0.Texture,ThreadID.x,ThreadID.y,TexelTransmittance);
}


/*! Like ApplyTransmittanceWeightedPrefixSumNonLinearRectificationMSM4() but 
   operates on textures storing six power moments in two textures, one with two and 
   one with four channels (useful for 16- or 32-bit quantization). The inputs are 
   supposed to be rectified already, so no coordinate conversion is performed.*/
void ApplyTransmittanceWeightedPrefixSumNonLinearRectificationMSM2_4(inout RWTexture2D<float2> OutFilteredShadowMap0,inout RWTexture2D<float4> OutFilteredShadowMap1,SSamplerTexturePair2D_float2 ShadowMapSampler0,SSamplerTexturePair2D ShadowMapSampler1,uint3 ThreadID,float ExtinctionCoefficient,float2 DistanceMinMax,float2 InclinationMinMax){
	uint Width,Height;
	OutFilteredShadowMap0.GetDimensions(Width,Height);
	float TexelTransmittance=ComputeTexelTransmittance(Width,ExtinctionCoefficient,DistanceMinMax,InclinationMinMax);
	// Different threads handle different textures
	if(ThreadID.y<Height){
		ApplyTransmittanceWeightedPrefixSum(OutFilteredShadowMap0,ShadowMapSampler0.Texture,ThreadID.x,ThreadID.y,TexelTransmittance);
	}
	else{
		ApplyTransmittanceWeightedPrefixSum(OutFilteredShadowMap1,ShadowMapSampler1.Texture,ThreadID.x,ThreadID.y-Height,TexelTransmittance);
	}
}


/*! Like ApplyTransmittanceWeightedPrefixSumNonLinearRectificationMSM2_4() but 
   stores the six moments in two three-channel textures (useful for 10-bit 
   quantization).*/
void ApplyTransmittanceWeightedPrefixSumNonLinearRectificationMSM3_3(inout RWTexture2D<float3> OutFilteredShadowMap0,inout RWTexture2D<float3> OutFilteredShadowMap1,SSamplerTexturePair2D_float3 ShadowMapSampler0,SSamplerTexturePair2D_float3 ShadowMapSampler1,uint3 ThreadID,float ExtinctionCoefficient,float2 DistanceMinMax,float2 InclinationMinMax){
	uint Width,Height;
	OutFilteredShadowMap0.GetDimensions(Width,Height);
	float TexelTransmittance=ComputeTexelTransmittance(Width,ExtinctionCoefficient,DistanceMinMax,InclinationMinMax);
	// Different threads handle different textures
	if(ThreadID.y<Height){
		ApplyTransmittanceWeightedPrefixSum(OutFilteredShadowMap0,ShadowMapSampler0.Texture,ThreadID.x,ThreadID.y,TexelTransmittance);
	}
	else{
		ApplyTransmittanceWeightedPrefixSum(OutFilteredShadowMap1,ShadowMapSampler1.Texture,ThreadID.x,ThreadID.y-Height,TexelTransmittance);
	}
}
