/*! This function computes the mesh to world space matrix for a vertex, which uses 
   skinning. Key frames are read from a texture.
  \param MeshToWorldSpaceSkinned The complete mesh to world space transformation.
  \param ObjectToWorldSpace The object to world space transformation which can be 
		 used to transform the entire object at once.
  \param Time The current time (in whichever unit used by the animation image).
  \param SkinWeights The skin weights of the considered vertex. The three given 
		 weights imply a third implicitly given weight (one minus their sum).
  \param SkinIndices The index of the transformation to be used for each of the 
		 provided skinning weights.
  \param AnimationTextureSampler A sampler for the texture containing the animation 
		 data (rotations, translations and scalings for each time and each possible 
		 skin index).
  \param AnimationTextureSize The resolution (xy) and reciprocal resolution (zw) of 
		 the texture provided by AnimationTextureSize.
  \param AnimationTextureTimeRange The beginning and end of the time interval 
		 covered by AnimationTextureSampler in x and y and 1.0/(y-x) in z.*/
void ComputeMeshToWorldSpaceSkinned(out float4x4 MeshToWorldSpaceSkinned,float4x4 ObjectToWorldSpace,float Time,float3 SkinWeights,int4 SkinIndices,SSamplerTexturePair2D AnimationTextureSampler,float4 AnimationTextureSize,float3 AnimationTextureTimeRange){
	// Construct a complete list of skin weights
	float4 AllSkinWeights=float4(SkinWeights.x,SkinWeights.y,SkinWeights.z,1.0f-dot(float3(1.0f,1.0f,1.0f),SkinWeights));
	// Compute the texture coordinate for the time-axis with proper wrapping and 
	// without going into no-mans land
	float WrappedRelativeTime=frac((Time-AnimationTextureTimeRange.x)*AnimationTextureTimeRange.z);
	float TimeCoordinate=lerp(0.5f*AnimationTextureTimeRange.z,1.0f-0.5f*AnimationTextureTimeRange.z,WrappedRelativeTime);
	// Prepare the mesh to object space matrix
	float4x4 MeshToObjectSpace=float4x4(0.0f,0.0f,0.0f,0.0f,0.0f,0.0f,0.0f,0.0f,0.0f,0.0f,0.0f,0.0f,0.0f,0.0f,0.0f,0.0f);
	[unroll] for(int i=0;i!=4;++i){
		// Compute the texture coordinate
		int iRotationRow=SkinIndices[i]*2;
		float2 RotationTexCoord=float2(TimeCoordinate,iRotationRow*AnimationTextureSize.w);
		RotationTexCoord.y+=0.5f*AnimationTextureSize.w;
		float2 TranslationScalingTexCoord=RotationTexCoord+float2(0.0f,AnimationTextureSize.w);
		// Sample the textures to obtain the coordinate frame
		float4 Rotation=normalize(AnimationTextureSampler.Texture.SampleLevel(AnimationTextureSampler.Sampler,RotationTexCoord,0.0f));
		float4 TranslationScaling=AnimationTextureSampler.Texture.SampleLevel(AnimationTextureSampler.Sampler,TranslationScalingTexCoord,0.0f);
		// Convert to a matrix and add to the weighted sum
		float xx=Rotation.x*Rotation.x;
		float xy=Rotation.x*Rotation.y;
		float xz=Rotation.x*Rotation.z;
		float xw=Rotation.x*Rotation.w;
		float yy=Rotation.y*Rotation.y;
		float yz=Rotation.y*Rotation.z;
		float yw=Rotation.y*Rotation.w;
		float zz=Rotation.z*Rotation.z;
		float zw=Rotation.z*Rotation.w;
		float Scaling=TranslationScaling.w;
		MeshToObjectSpace+=AllSkinWeights[i]*float4x4(
			Scaling*(1.0f-2.0f*(yy+zz)),2.0f*Scaling*(xy+zw),2.0f*Scaling*(xz-yw),0.0f,
			2.0f*Scaling*(xy-zw),Scaling*(1.0f-2.0f*(xx+zz)),2.0f*Scaling*(yz+xw),0.0f,
			2.0f*Scaling*(xz+yw),2.0f*Scaling*(yz-xw),Scaling*(1.0f-2.0f*(xx+yy)),0.0f,
			TranslationScaling.x,TranslationScaling.y,TranslationScaling.z,1.0f);
	}
	// Go to world space
	MeshToWorldSpaceSkinned=mul(MeshToObjectSpace,ObjectToWorldSpace);
}
