/*! This function interpolates lineary between the given two positions using the 
   given lerp factor.*/
void ApplyTweening(out float3 MeshSpacePositionTweened,float3 MeshSpacePosition,float3 MeshSpacePosition2,float TweeningFactor){
	MeshSpacePositionTweened=lerp(MeshSpacePosition,MeshSpacePosition2,TweeningFactor);
}

/*! This function uses the given mesh to world space transformation matrix to get a 
   vertex position from mesh space to world space.*/
void ApplyMeshToWorldSpace(out float3 WorldSpacePosition,float3 MeshSpacePositionCombined,float4x4 MeshToWorldSpace){
	WorldSpacePosition=mul(float4(MeshSpacePositionCombined,1.0f),MeshToWorldSpace).xyz;
}

/*! This function interpolates lineary between the given two normals using the 
   given lerp factor.*/
void ApplyTweeningNormal(out float3 MeshSpaceNormalTweened,float3 MeshSpaceNormal,float3 MeshSpaceNormal2,float TweeningFactor){
	MeshSpaceNormalTweened=lerp(MeshSpaceNormal,MeshSpaceNormal2,TweeningFactor);
}

/*! This function uses the given mesh to world space transformation matrix to get a 
   vertex normal from mesh space to world space.*/
void ApplyMeshToWorldSpaceNormal(out float3 WorldSpaceNormal,float3 MeshSpaceNormalCombined,float4x4 MeshToWorldSpace){
	float3x3 MeshToWorldSpaceDirection=(float3x3)MeshToWorldSpace;
	WorldSpaceNormal=normalize(mul(MeshSpaceNormalCombined,MeshToWorldSpaceDirection)*determinant(MeshToWorldSpaceDirection));
}

/*! This function uses the given world to projection space transformation matrix to 
   get a vertex position from world space to projection space.*/
void ApplyWorldToProjectionSpace(out float4 ProjectionSpacePosition,float3 WorldSpacePosition,float4x4 WorldToProjectionSpace){
	ProjectionSpacePosition=mul(float4(WorldSpacePosition,1.0f),WorldToProjectionSpace);
}


/*! This function prepares the information, which is needed to compute the tangent to 
   world space transformation in the pixel shader. Namely it computes the normalized 
   world space tangent and the bitangent factor. It is meant to be used in the 
   vertex shader.*/
void PrepareTangentToWorldSpace(out float3 OutWorldSpaceTangent,out float OutBitangentFactor,float4x4 MeshToWorldSpace,float4 TangentSpaceData){
	OutWorldSpaceTangent=normalize(mul(TangentSpaceData.xyz,(float3x3)MeshToWorldSpace));
	OutBitangentFactor=TangentSpaceData.w;
}

/*! This function constructs the tangent to world space transformation as 3x3 matrix 
   using the data computed by PrepareTangentToWorldSpace(...). Due to interpolation 
   the returned matrix is not necessarily orthogonal. The function is meant to be 
   used in the pixel shader.*/
void ComputeTangentToWorldSpace(out float3x3 OutTangentToWorldSpace,float3 WorldSpaceNormal,float3 WorldSpaceTangent,float BitangentFactor){
	// Compute the bitangent
	float3 WorldSpaceBitangent=cross(WorldSpaceNormal,WorldSpaceTangent)*BitangentFactor;
	// Construct the matrix
	OutTangentToWorldSpace._11=WorldSpaceTangent.x;
	OutTangentToWorldSpace._21=WorldSpaceTangent.y;
	OutTangentToWorldSpace._31=WorldSpaceTangent.z;
	OutTangentToWorldSpace._12=WorldSpaceBitangent.x;
	OutTangentToWorldSpace._22=WorldSpaceBitangent.y;
	OutTangentToWorldSpace._32=WorldSpaceBitangent.z;
	OutTangentToWorldSpace._13=WorldSpaceNormal.x;
	OutTangentToWorldSpace._23=WorldSpaceNormal.y;
	OutTangentToWorldSpace._33=WorldSpaceNormal.z;
}

/*! This function transforms a given tangent space normal from tangent space to world 
   space and normalizes it afterwards.*/
void TangentSpaceNormalToWorldSpace(out float3 OutWorldSpaceNormal,float3 TangentSpaceNormal,float3x3 TangentToWorldSpace){
	OutWorldSpaceNormal=normalize(mul(TangentToWorldSpace,TangentSpaceNormal));
}


/*! This function applies a texture coordinate dependent offset as it is commonly 
   used for billboards to the vertices to create a quad of constant size.*/
void BillboardOffsetConstantSize(out float3 WorldSpacePositionOffset,float3 WorldSpacePosition,float2 BillboardTexCoord,float2 BillboardOrigin,float3 BillboardScaledRightVector,float3 BillboardScaledUpVector){
	float2 BillboardOffset=BillboardTexCoord-BillboardOrigin;
	WorldSpacePositionOffset=WorldSpacePosition;
	WorldSpacePositionOffset+=BillboardScaledRightVector*BillboardOffset.x;
	WorldSpacePositionOffset+=BillboardScaledUpVector*-BillboardOffset.y;
}


/*! This function applies a texture coordinate dependent offset as it is commonly 
   used for billboards to the vertices to create a quad of variable size.*/
void BillboardOffsetVariableSize(out float3 WorldSpacePositionOffset,float3 WorldSpacePosition,float2 BillboardTexCoord,float2 BillboardSize,float2 BillboardOrigin,float3 BillboardRightVector,float3 BillboardUpVector){
	float2 BillboardOffset=(BillboardTexCoord-BillboardOrigin)*BillboardSize;
	WorldSpacePositionOffset=WorldSpacePosition;
	WorldSpacePositionOffset+=BillboardRightVector*BillboardOffset.x;
	WorldSpacePositionOffset+=BillboardUpVector*-BillboardOffset.y;
}


/*! This function applies a texture coordinate dependent offset as it is commonly 
   used for billboards to the vertices to create a quad of variable size, which is 
   rotated through a variable radian value.*/
void BillboardOffsetVariableSizeAndRotation(out float3 WorldSpacePositionOffset,float3 WorldSpacePosition,float2 BillboardTexCoord,float2 BillboardSize,float BillboardRotation,float2 BillboardOrigin,float3 BillboardRightVector,float3 BillboardUpVector){
	float2 SineCosine;
	sincos(BillboardRotation,SineCosine.x,SineCosine.y);
	float2 BillboardOffset=(BillboardTexCoord-BillboardOrigin)*BillboardSize;
	WorldSpacePositionOffset=WorldSpacePosition;
	WorldSpacePositionOffset+=(BillboardRightVector*SineCosine.y+BillboardUpVector*SineCosine.x)*BillboardOffset.x;
	WorldSpacePositionOffset+=(BillboardUpVector*SineCosine.y-BillboardRightVector*SineCosine.x)*-BillboardOffset.y;
}


/*! This function sets the z-coordinate of the given projection space position to its 
   w-coordinate and outputs the result. It can be used to get surfaces with a z-
   buffer value of 1.0 (i.e. background surfaces).*/
void MaximizeProjectionZ(out float4 ProjectionSpacePositionBackground,float4 ProjectionSpacePosition){
	ProjectionSpacePositionBackground.xyw=ProjectionSpacePosition.xyw;
	ProjectionSpacePositionBackground.z=ProjectionSpacePosition.w*0.999f;
}
