/*! This function uses the red component of the given fragment color multiplied by 
   the alpha component of the given text color as alpha and overwrites the color 
   components with the color components of the text color.*/
void ComputeTextColor(out float4 OutFragmentColor,float4 FragmentColor,float4 TextColor=float4(1.0f,1.0f,1.0f,1.0f)){
	OutFragmentColor.a=FragmentColor.r*TextColor.a;
	OutFragmentColor.rgb=TextColor.rgb;
}

/*! This function divides the first given vector by the second given vector 
   component-wise. This can be used to bring a sprite-space position into the range 
   from 0.0 to 1.0.*/
void NormalizeSpriteSpacePosition(out float2 OutSpriteSpacePosition,float2 Position,float2 Divisor){
	OutSpriteSpacePosition.x=Position.x/Divisor.x;
	OutSpriteSpacePosition.y=Position.y/Divisor.y;
}

/*! This function returns the additive inverse of the given 2D vector. This can be 
   interpreted as inverse of a translation.*/
void ComputeInverseTranslation(out float2 OutInverseTranslation,float2 Translation){
	OutInverseTranslation=-Translation;
}

/*! This function applies the given translation to the given position and outputs the 
   result. In other words it performs per-component addition.*/
void Apply2DTranslation(out float2 OutTranslatedPosition,float2 Position,float2 Translation){
	OutTranslatedPosition=Position+Translation;
}

/*! This function applies a 2D transformation (rotation, scaling and translation) to 
   a given 2D vector and outputs the result.*/
void Apply2DTransformation(out float2 OutTransformedPosition,float2 Position,float3x2 Transformation){
	OutTransformedPosition=mul(float3(Position,1.0f),Transformation).xy;
}

/*! This function applies a 3D transformation (rotation, scaling and translation) to 
   a position in the x-y-plane and outputs the result.*/
void Apply3DTransformation(out float4 OutTransformedPosition,float2 Position,float4x4 Transformation){
	OutTransformedPosition=mul(float4(Position,0.0f,1.0f),Transformation);
}

/*! This function first applies a 2D scaling and then a 2D translation to a given 
   position and returns the result. The scaling is ScalingAndTranslation.xy, the 
   translation is ScalingAndTranslation.zw.*/
void Apply2DScalingAndTranslation(out float2 OutTransformedPosition,float2 Position,float4 ScalingAndTranslation){
	OutTransformedPosition=Position.xy*ScalingAndTranslation.xy+ScalingAndTranslation.zw;
}


/*! This function uses the given data (same names as data bindings of 
   C2DLibraryDefinition) to construct a sprite to viewport space transformation.*/
void ComputeSpriteToViewportSpace(out float3x2 OutSpriteToViewportSpace,float2 ViewportSpaceSpritePosition,float SpriteRotation,float2 SpriteSpaceCenter,float2 SpriteSize){
	// Construct the rotation and scaling part
	float RotationSin,RotationCos;
	sincos(SpriteRotation,RotationSin,RotationCos);
	OutSpriteToViewportSpace._11=RotationCos*SpriteSize.x;
	OutSpriteToViewportSpace._12=-RotationSin*SpriteSize.x;
	OutSpriteToViewportSpace._21=RotationSin*SpriteSize.y;
	OutSpriteToViewportSpace._22=RotationCos*SpriteSize.y;
	// Add the translation
	float2 ViewportCenterOffest=mul(SpriteSpaceCenter,float2x2(OutSpriteToViewportSpace._11,OutSpriteToViewportSpace._12,OutSpriteToViewportSpace._21,OutSpriteToViewportSpace._22));
	ViewportSpaceSpritePosition-=ViewportCenterOffest;
	OutSpriteToViewportSpace._31=ViewportSpaceSpritePosition.x;
	OutSpriteToViewportSpace._32=ViewportSpaceSpritePosition.y;
}


/*! This function constructs a matrix, which transforms viewport coordinates for a 
   viewport of given size into projection space coordinates.*/
void ComputeViewportToProjectionSpace(out float3x4 OutViewportToProjectionSpace,float2 ViewportSize){
	float2 InvViewportSize=float2(1.0f/ViewportSize.x,1.0f/ViewportSize.y);
	OutViewportToProjectionSpace._11=2.0f*InvViewportSize.x;
	OutViewportToProjectionSpace._12=0.0f;
	OutViewportToProjectionSpace._13=0.0f;
	OutViewportToProjectionSpace._14=0.0f;

	OutViewportToProjectionSpace._21=0.0f;
	OutViewportToProjectionSpace._22=-2.0f*InvViewportSize.y;
	OutViewportToProjectionSpace._23=0.0f;
	OutViewportToProjectionSpace._24=0.0f;

	OutViewportToProjectionSpace._31=-1.0f;
	OutViewportToProjectionSpace._32=1.0f;
	OutViewportToProjectionSpace._33=0.0f;
	OutViewportToProjectionSpace._34=1.0f;
}


/*! This function applies a given viewport to projection space transform to a 
   viewport space vector and outputs the result. The parameter names are compliant 
   with with the data bindings of C2DLibraryDefinition.*/
void Apply2DProjectionSpaceTransform(out float4 OutProjectionSpacePosition,float2 ViewportSpacePosition,float3x4 ViewportToProjectionSpace){
	OutProjectionSpacePosition=mul(float3(ViewportSpacePosition,1.0f),ViewportToProjectionSpace);
}


/*! Given a planar coordinate this function computes a corresponding projection 
   space position. 
  \param Position A two-dimensional coordinate identifying a location in the 
		 viewport, which is to be represented by the output projection space 
		 position. (0.0,0.0) identifies the left top of the viewport (1.0,1.0) 
		 identifies the right bottom and the x-coordiante corresponds to the 
		 horizontal axis.*/
void ComputePlanarProjectionSpacePosition(out float4 OutProjectionSpacePosition,float2 Position){
	OutProjectionSpacePosition=float4(2.0f*Position.x-1.0f,-2.0f*Position.y+1.0f,0.5f,1.0f);
}


/*! This function computes positions of vertices of a graph.
  \param OutPosition The offset from the origin of the coordinate system in pixels.
  \param XCoordinate,YCoordinate The coordinates of a point on the graph.
  \param DomainTransform The transformation information provided by 
		 CDomainTransform.
  \param YOffset A vertical offset in pixels.*/
void ComputeGraphPosition(out float2 OutPosition,float XCoordinate,float YCoordinate,float2 DomainTransform,float YOffset){
	OutPosition=float2(XCoordinate*DomainTransform.x,YCoordinate*DomainTransform.y+YOffset);
}

/*! Describes a vertex as input to the geometry shader.*/
struct SGraphInputVertex{
	/*! The x-coordinate of the vertex in domain coordinates.*/
	float XCoordinate;
	/*! The y-coordinate of the vertex in domain coordinates.*/
	float YCoordinate;
	/*! The line width in pixels at this vertex divided by two.*/
	float HalfLineWidth;
};
/*! Describes a vertex as output by the geometry shader.*/
struct SGraphOutputVertex{
	/*! The position of the vertex in projection space.*/
	float4 ProjectionSpacePosition;
};

/*! This function generates primitives for a single line segment making up a graph.
  \param OutTriangleStream The output stream to which triangles for the graph are 
		 pushed.
  \param pLineVertex The vertices of the current line segment.
  \param DomainTransform The transformation information provided by 
		 CDomainTransform.
  \param Divisor The size of the domain in pixels.
  \param SpriteToViewportSpace The transformation needed to go from sprite to 
		 viewport space.
  \param ViewportToProjectionSpace The transformation needed to go from viewport to 
		 projection space.*/
void GenerateGraphQuad(inout TriangleStream<SGraphOutputVertex> OutTriangleStream,SGraphInputVertex pLineVertex[2],float2 DomainTransform,float2 Divisor,float3x2 SpriteToViewportSpace,float3x4 ViewportToProjectionSpace){
	// Get to sprite space in pixels
	float HalfLineWidth=pLineVertex[0].HalfLineWidth;
	float2 pLineSpriteSpacePosition[2]={
		float2(pLineVertex[0].XCoordinate,pLineVertex[0].YCoordinate)*DomainTransform,
		float2(pLineVertex[1].XCoordinate,pLineVertex[1].YCoordinate)*DomainTransform
	};
	// Generate the triangle strip
	float2 LineVector=pLineSpriteSpacePosition[1]-pLineSpriteSpacePosition[0];
	float2 OrthonormalVector=normalize(float2(LineVector.y,-LineVector.x));
	float2 pSpriteSpacePosition[4]={
		pLineSpriteSpacePosition[0]-OrthonormalVector*HalfLineWidth,
		pLineSpriteSpacePosition[0]+OrthonormalVector*HalfLineWidth,
		pLineSpriteSpacePosition[1]-OrthonormalVector*HalfLineWidth,
		pLineSpriteSpacePosition[1]+OrthonormalVector*HalfLineWidth
	};
	// Transform to projection space and generate the quad
	[unroll] for(int i=0;i!=4;++i){
		float2 SpriteSpacePosition=pSpriteSpacePosition[i];
		// Go to normalized sprite space
		float2 NormalizedSpriteSpacePosition=float2(SpriteSpacePosition.x/Divisor.x,SpriteSpacePosition.y/Divisor.y);
		// Go to viewport space
		float2 ViewportSpacePosition=mul(float3(NormalizedSpriteSpacePosition,1.0f),SpriteToViewportSpace);
		// Go to projection space
		float4 ProjectionSpacePosition=mul(float3(ViewportSpacePosition,1.0f),ViewportToProjectionSpace);
		// Output the vertex
		SGraphOutputVertex NewVertex;
		NewVertex.ProjectionSpacePosition=ProjectionSpacePosition;
		OutTriangleStream.Append(NewVertex);
	}
}
